"""Knowledge resource."""

import dataclasses
from typing import TYPE_CHECKING, Any, List, Optional, Type

import cachetools

from dbgpt.util.cache_utils import cached

from .base import Resource, ResourceParameters, ResourceType

if TYPE_CHECKING:
    from dbgpt.core import Chunk
    from dbgpt.rag.retriever.base import BaseRetriever
    from dbgpt.storage.vector_store.filters import MetadataFilters


@dataclasses.dataclass
class RetrieverResourceParameters(ResourceParameters):
    """Retriever resource parameters."""

    pass


class RetrieverResource(Resource[ResourceParameters]):
    """Retriever resource.

    Retrieve knowledge chunks from a retriever.
    """

    def __init__(self, name: str, retriever: "BaseRetriever"):
        """Create a new RetrieverResource."""
        self._name = name
        self._retriever = retriever

    @property
    def name(self) -> str:
        """Return the resource name."""
        return self._name

    @property
    def retriever(self) -> "BaseRetriever":
        """Return the retriever."""
        return self._retriever

    @classmethod
    def type(cls) -> ResourceType:
        """Return the resource type."""
        return ResourceType.Knowledge

    @classmethod
    def resource_parameters_class(cls) -> Type[ResourceParameters]:
        """Return the resource parameters class."""
        return RetrieverResourceParameters

    @cached(cachetools.TTLCache(maxsize=100, ttl=10))
    async def get_prompt(
        self,
        *,
        lang: str = "en",
        prompt_type: str = "default",
        question: Optional[str] = None,
        resource_name: Optional[str] = None,
        **kwargs
    ) -> str:
        """Get the prompt for the resource."""
        if not question:
            raise ValueError("Question is required for knowledge resource.")
        chunks = await self.retrieve(question)
        content = "\n".join([chunk.content for chunk in chunks])
        prompt_template = "known information: {content}"
        prompt_template_zh = "已知信息: {content}"
        if lang == "en":
            return prompt_template.format(content=content)
        return prompt_template_zh.format(content=content)

    async def async_execute(
        self, *args, resource_name: Optional[str] = None, **kwargs
    ) -> Any:
        """Execute the resource asynchronously."""
        return await self.retrieve(*args, **kwargs)

    async def retrieve(
        self, query: str, filters: Optional["MetadataFilters"] = None
    ) -> List["Chunk"]:
        """Retrieve knowledge chunks.

        Args:
            query (str): query text.
            filters: (Optional[MetadataFilters]) metadata filters.

        Returns:
            List[Chunk]: list of chunks
        """
        return await self.retriever.aretrieve(query, filters)



"""Resources for the agent."""

import dataclasses
import json
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar, cast

from dbgpt._private.pydantic import BaseModel, model_to_dict
from dbgpt.util.parameter_utils import BaseParameters, _get_parameter_descriptions

P = TypeVar("P", bound="ResourceParameters")
T = TypeVar("T", bound="Resource")


class ResourceType(str, Enum):
    """Resource type enumeration."""

    DB = "database"
    Knowledge = "knowledge"
    Internet = "internet"
    Tool = "tool"
    Plugin = "plugin"
    TextFile = "text_file"
    ExcelFile = "excel_file"
    ImageFile = "image_file"
    AWELFlow = "awel_flow"
    # Resource type for resource pack
    Pack = "pack"


@dataclasses.dataclass
class ResourceParameters(BaseParameters):
    """Resource parameters class.

    It defines the parameters for building a resource.
    """

    name: str = dataclasses.field(metadata={"help": "Resource name", "tags": "fixed"})

    @classmethod
    def _resource_version(cls) -> str:
        """Return the resource version."""
        return "v2"

    @classmethod
    def to_configurations(
        cls, parameters: Type["ResourceParameters"], version: Optional[str] = None
    ) -> Any:
        """Convert the parameters to configurations."""
        return _get_parameter_descriptions(parameters)


class Resource(ABC, Generic[P]):
    """Resource for the agent."""

    @classmethod
    @abstractmethod
    def type(cls) -> ResourceType:
        """Return the resource type."""

    @classmethod
    def type_alias(cls) -> str:
        """Return the resource type alias."""
        return cls.type().value

    @property
    @abstractmethod
    def name(self) -> str:
        """Return the resource name."""

    @classmethod
    def resource_parameters_class(cls) -> Type[P]:
        """Return the parameters class."""
        return ResourceParameters

    def prefer_resource_parameters_class(self) -> Type[P]:
        """Return the parameters class.

        You can override this method to return a different parameters class.
        It will be used to initialize the resource with parameters.
        """
        return self.resource_parameters_class()

    def initialize_with_parameters(self, resource_parameters: P):
        """Initialize the resource with parameters."""
        pass

    def preload_resource(self):
        """Preload the resource."""
        pass

    @classmethod
    def from_resource(
        cls: Type[T],
        resource: Optional["Resource"],
        expected_type: Optional[ResourceType] = None,
    ) -> List[T]:
        """Create a resource from another resource.

        Another resource can be a pack or a single resource, if it is a pack, it will
        return all resources which type is the same as the current resource.

        Args:
            resource(Resource): The resource.
            expected_type(ResourceType): The expected resource type.
        Returns:
            List[Resource]: The resources.
        """
        if not resource:
            return []
        typed_resources = []
        for r in resource.get_resource_by_type(expected_type or cls.type()):
            typed_resources.append(cast(T, r))
        return typed_resources

    @abstractmethod
    async def get_prompt(
        self,
        *,
        lang: str = "en",
        prompt_type: str = "default",
        question: Optional[str] = None,
        resource_name: Optional[str] = None,
        **kwargs,
    ) -> str:
        """Get the prompt.

        Args:
            lang(str): The language.
            prompt_type(str): The prompt type.
            question(str): The question.
            resource_name(str): The resource name, just for the pack, it will be used
                to select specific resource in the pack.
        """

    def execute(self, *args, resource_name: Optional[str] = None, **kwargs) -> Any:
        """Execute the resource."""
        raise NotImplementedError

    async def async_execute(
        self, *args, resource_name: Optional[str] = None, **kwargs
    ) -> Any:
        """Execute the resource asynchronously."""
        raise NotImplementedError

    @property
    def is_async(self) -> bool:
        """Return whether the resource is asynchronous."""
        return False

    @property
    def is_pack(self) -> bool:
        """Return whether the resource is a pack."""
        return False

    @property
    def sub_resources(self) -> List["Resource"]:
        """Return the resources."""
        if not self.is_pack:
            raise ValueError("The resource is not a pack, no sub-resources.")
        return []

    def get_resource_by_type(self, resource_type: ResourceType) -> List["Resource"]:
        """Get resources by type.

        If the resource is a pack, it will search the sub-resources. Otherwise, it will
        return itself if the type matches.

        Args:
            resource_type(ResourceType): The resource type.

        Returns:
            List[Resource]: The resources.
        """
        if not self.is_pack:
            if self.type() == resource_type:
                return [self]
            else:
                return []
        resources = []
        for resource in self.sub_resources:
            if resource.type() == resource_type:
                resources.append(resource)
        return resources


class AgentResource(BaseModel):
    """Agent resource class."""

    type: str
    name: str
    value: str
    is_dynamic: bool = (
        False  # Is the current resource predefined or dynamically passed in?
    )

    def resource_prompt_template(self, **kwargs) -> str:
        """Get the resource prompt template."""
        return "{data_type}  --{data_introduce}"

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> Optional["AgentResource"]:
        """Create an AgentResource object from a dictionary."""
        if d is None:
            return None
        return AgentResource(
            type=d.get("type"),
            name=d.get("name"),
            introduce=d.get("introduce"),
            value=d.get("value", None),
            is_dynamic=d.get("is_dynamic", False),
            parameters=d.get("parameters", None),
        )

    @staticmethod
    def from_json_list_str(d: Optional[str]) -> Optional[List["AgentResource"]]:
        """Create a list of AgentResource objects from a json string."""
        if d is None:
            return None
        try:
            json_array = json.loads(d)
        except Exception:
            raise ValueError(f"Illegal AgentResource json string！{d}")
        if not isinstance(json_array, list):
            raise ValueError(f"Illegal AgentResource json string！{d}")
        json_list = []
        for item in json_array:
            r = AgentResource.from_dict(item)
            if r:
                json_list.append(r)
        return json_list

    def to_dict(self) -> Dict[str, Any]:
        """Convert the AgentResource object to a dictionary."""
        temp = model_to_dict(self)
        for field, value in temp.items():
            if isinstance(value, Enum):
                temp[field] = value.value
        return temp



class LLMChatError(Exception):
    """
    llm conversation result generates exception
    """

    def __init__(self, message="LLM Chat Generrate Error!", original_exception=None):
        super().__init__(message)
        self.message = message
        self.original_exception = original_exception

    def __str__(self):
        if self.original_exception:
            # 返回自定义异常信息和原始异常信息
            return f"{self.message}({self.original_exception})"
        else:
            return self.message



"""PDF Knowledge."""
from typing import Any, Dict, List, Optional, Union

from dbgpt.core import Document
from dbgpt.rag.knowledge.base import (
    ChunkStrategy,
    DocumentType,
    Knowledge,
    KnowledgeType,
)


class PDFKnowledge(Knowledge):
    """PDF Knowledge."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        knowledge_type: KnowledgeType = KnowledgeType.DOCUMENT,
        loader: Optional[Any] = None,
        language: Optional[str] = "zh",
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
        **kwargs: Any,
    ) -> None:
        """Create PDF Knowledge with Knowledge arguments.

        Args:
            file_path(str,  optional): file path
            knowledge_type(KnowledgeType, optional): knowledge type
            loader(Any, optional): loader
            language(str, optional): language
        """
        super().__init__(
            path=file_path,
            knowledge_type=knowledge_type,
            data_loader=loader,
            metadata=metadata,
            **kwargs,
        )
        self._language = language

    def _load(self) -> List[Document]:
        """Load pdf document from loader."""
        if self._loader:
            documents = self._loader.load()
        else:
            import pypdf

            pages = []
            documents = []
            if not self._path:
                raise ValueError("file path is required")
            with open(self._path, "rb") as file:
                reader = pypdf.PdfReader(file)
                for page_num in range(len(reader.pages)):
                    _page = reader.pages[page_num]
                    pages.append((_page.extract_text(), page_num))

            # cleaned_pages = []
            for page, page_num in pages:
                lines = page.splitlines()

                cleaned_lines = []
                for line in lines:
                    if self._language == "en":
                        words = list(line)  # noqa: F841
                    else:
                        words = line.split()  # noqa: F841
                    cleaned_lines.append(line)
                page = "\n".join(cleaned_lines)
                # cleaned_pages.append(page)
                metadata = {"source": self._path, "page": page_num}
                if self._metadata:
                    metadata.update(self._metadata)  # type: ignore
                # text = "\f".join(cleaned_pages)
                document = Document(content=page, metadata=metadata)
                documents.append(document)
            return documents
        return [Document.langchain2doc(lc_document) for lc_document in documents]

    @classmethod
    def support_chunk_strategy(cls) -> List[ChunkStrategy]:
        """Return support chunk strategy."""
        return [
            ChunkStrategy.CHUNK_BY_SIZE,
            ChunkStrategy.CHUNK_BY_PAGE,
            ChunkStrategy.CHUNK_BY_SEPARATOR,
        ]

    @classmethod
    def default_chunk_strategy(cls) -> ChunkStrategy:
        """Return default chunk strategy."""
        return ChunkStrategy.CHUNK_BY_SIZE

    @classmethod
    def type(cls) -> KnowledgeType:
        """Return knowledge type."""
        return KnowledgeType.DOCUMENT

    @classmethod
    def document_type(cls) -> DocumentType:
        """Document type of PDF."""
        return DocumentType.PDF



"""Re-rank embeddings."""

from typing import Any, Dict, List, Optional, cast

import aiohttp
import requests

from dbgpt._private.pydantic import EXTRA_FORBID, BaseModel, ConfigDict, Field
from dbgpt.core import RerankEmbeddings


class CrossEncoderRerankEmbeddings(BaseModel, RerankEmbeddings):
    """CrossEncoder Rerank Embeddings."""

    model_config = ConfigDict(extra=EXTRA_FORBID, protected_namespaces=())

    client: Any  #: :meta private:
    model_name: str = "BAAI/bge-reranker-base"
    max_length: Optional[int] = None
    """Max length for input sequences. Longer sequences will be truncated. If None, max
        length of the model will be used"""
    """Model name to use."""
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Keyword arguments to pass to the model."""

    def __init__(self, **kwargs: Any):
        """Initialize the sentence_transformer."""
        try:
            from sentence_transformers import CrossEncoder
        except ImportError:
            raise ImportError(
                "please `pip install sentence-transformers`",
            )

        kwargs["client"] = CrossEncoder(
            kwargs.get("model_name"),
            max_length=kwargs.get("max_length"),
            **kwargs.get("model_kwargs"),
        )
        super().__init__(**kwargs)

    def predict(self, query: str, candidates: List[str]) -> List[float]:
        """Predict the rank scores of the candidates.

        Args:
            query: The query text.
            candidates: The list of candidate texts.

        Returns:
            List[float]: The rank scores of the candidates.
        """
        from sentence_transformers import CrossEncoder

        query_content_pairs = [[query, candidate] for candidate in candidates]
        _model = cast(CrossEncoder, self.client)
        rank_scores = _model.predict(sentences=query_content_pairs)
        return rank_scores.tolist()


class OpenAPIRerankEmbeddings(BaseModel, RerankEmbeddings):
    """OpenAPI Rerank Embeddings."""

    model_config = ConfigDict(arbitrary_types_allowed=True, protected_namespaces=())

    api_url: str = Field(
        default="http://localhost:8100/v1/beta/relevance",
        description="The URL of the embeddings API.",
    )
    api_key: Optional[str] = Field(
        default=None, description="The API key for the embeddings API."
    )
    model_name: str = Field(
        default="bge-reranker-base", description="The name of the model to use."
    )
    timeout: int = Field(
        default=60, description="The timeout for the request in seconds."
    )

    session: Optional[requests.Session] = None

    def __init__(self, **kwargs):
        """Initialize the OpenAPIEmbeddings."""
        try:
            import requests
        except ImportError:
            raise ValueError(
                "The requests python package is not installed. "
                "Please install it with `pip install requests`"
            )
        if "session" not in kwargs:  # noqa: SIM401
            session = requests.Session()
        else:
            session = kwargs["session"]
        api_key = kwargs.get("api_key")
        if api_key:
            session.headers.update({"Authorization": f"Bearer {api_key}"})
        kwargs["session"] = session
        super().__init__(**kwargs)

    def predict(self, query: str, candidates: List[str]) -> List[float]:
        """Predict the rank scores of the candidates.

        Args:
            query: The query text.
            candidates: The list of candidate texts.

        Returns:
            List[float]: The rank scores of the candidates.
        """
        if not candidates:
            return []
        data = {"model": self.model_name, "query": query, "documents": candidates}
        response = self.session.post(  # type: ignore
            self.api_url, json=data, timeout=self.timeout
        )
        response.raise_for_status()
        return response.json()["data"]

    async def apredict(self, query: str, candidates: List[str]) -> List[float]:
        """Predict the rank scores of the candidates asynchronously."""
        headers = {"Authorization": f"Bearer {self.api_key}"}
        async with aiohttp.ClientSession(
            headers=headers, timeout=aiohttp.ClientTimeout(total=self.timeout)
        ) as session:
            data = {"model": self.model_name, "query": query, "documents": candidates}
            async with session.post(self.api_url, json=data) as resp:
                resp.raise_for_status()
                response_data = await resp.json()
                if "data" not in response_data:
                    raise RuntimeError(response_data["detail"])
                return response_data["data"]



"""The module of stream operator."""
from abc import ABC, abstractmethod
from typing import AsyncIterator, Generic

from ..dag.base import DAGContext
from ..task.base import IN, OUT, TaskContext, TaskOutput
from .base import BaseOperator


class StreamifyAbsOperator(BaseOperator[OUT], ABC, Generic[IN, OUT]):
    """An abstract operator that converts a value of IN to an AsyncIterator[OUT]."""

    streaming_operator = True

    async def _do_run(self, dag_ctx: DAGContext) -> TaskOutput[OUT]:
        curr_task_ctx: TaskContext[OUT] = dag_ctx.current_task_context
        call_data = curr_task_ctx.call_data
        if call_data:
            wrapped_call_data = await curr_task_ctx._call_data_to_output()
            if not wrapped_call_data:
                raise ValueError(
                    f"task {curr_task_ctx.task_id} MapDAGNode expects wrapped_call_data"
                )
            output = await wrapped_call_data.streamify(self.streamify)
            curr_task_ctx.set_task_output(output)
            return output
        output = await curr_task_ctx.task_input.parent_outputs[0].task_output.streamify(
            self.streamify
        )
        curr_task_ctx.set_task_output(output)
        return output

    @abstractmethod
    async def streamify(self, input_value: IN) -> AsyncIterator[OUT]:
        """Convert a value of IN to an AsyncIterator[OUT].

        Args:
            input_value (IN): The data of parent operator's output

        Examples:
            .. code-block:: python

                class MyStreamOperator(StreamifyAbsOperator[int, int]):
                    async def streamify(self, input_value: int) -> AsyncIterator[int]:
                        for i in range(input_value):
                            yield i

        """


class UnstreamifyAbsOperator(BaseOperator[OUT], Generic[IN, OUT]):
    """An abstract operator that converts a value of AsyncIterator[IN] to an OUT."""

    async def _do_run(self, dag_ctx: DAGContext) -> TaskOutput[OUT]:
        curr_task_ctx: TaskContext[OUT] = dag_ctx.current_task_context
        output: TaskOutput[OUT] = await curr_task_ctx.task_input.parent_outputs[
            0
        ].task_output.unstreamify(self.unstreamify)
        curr_task_ctx.set_task_output(output)
        return output

    @abstractmethod
    async def unstreamify(self, input_value: AsyncIterator[IN]) -> OUT:
        """Convert a value of AsyncIterator[IN] to an OUT.

        Args:
            input_value (AsyncIterator[IN])): The data of parent operator's output

        Example:
            .. code-block:: python

                class MyUnstreamOperator(UnstreamifyAbsOperator[int, int]):
                    async def unstreamify(self, input_value: AsyncIterator[int]) -> int:
                        value_cnt = 0
                        async for v in input_value:
                            value_cnt += 1
                        return value_cnt
        """


class TransformStreamAbsOperator(BaseOperator[OUT], Generic[IN, OUT]):
    """Streaming to other streaming data.

    An abstract operator that transforms a value of
    AsyncIterator[IN] to another AsyncIterator[OUT].
    """

    streaming_operator = True

    async def _do_run(self, dag_ctx: DAGContext) -> TaskOutput[OUT]:
        curr_task_ctx: TaskContext[OUT] = dag_ctx.current_task_context
        output: TaskOutput[OUT] = await curr_task_ctx.task_input.parent_outputs[
            0
        ].task_output.transform_stream(self.transform_stream)

        curr_task_ctx.set_task_output(output)
        return output

    @abstractmethod
    async def transform_stream(
        self, input_value: AsyncIterator[IN]
    ) -> AsyncIterator[OUT]:
        """Transform an AsyncIterator[IN] to another AsyncIterator[OUT].

        Args:
            input_value (AsyncIterator[IN])): The data of parent operator's output

        Examples:
            .. code-block:: python

                class MyTransformStreamOperator(TransformStreamAbsOperator[int, int]):
                    async def unstreamify(
                        self, input_value: AsyncIterator[int]
                    ) -> AsyncIterator[int]:
                        async for v in input_value:
                            yield v + 1
        """



"""Doris connector."""
from typing import Any, Dict, Iterable, List, Optional, Tuple, cast
from urllib.parse import quote
from urllib.parse import quote_plus as urlquote

from sqlalchemy import text

from .base import RDBMSConnector


class DorisConnector(RDBMSConnector):
    """Doris connector."""

    driver = "doris"
    db_type = "doris"
    db_dialect = "doris"

    @classmethod
    def from_uri_db(
        cls,
        host: str,
        port: int,
        user: str,
        pwd: str,
        db_name: str,
        engine_args: Optional[dict] = None,
        **kwargs: Any,
    ) -> "DorisConnector":
        """Create a new DorisConnector from host, port, user, pwd, db_name."""
        db_url: str = (
            f"{cls.driver}://{quote(user)}:{urlquote(pwd)}@{host}:{str(port)}/{db_name}"
        )
        return cast(DorisConnector, cls.from_uri(db_url, engine_args, **kwargs))

    def _sync_tables_from_db(self) -> Iterable[str]:
        table_results = self.get_session().execute(
            text(
                "SELECT TABLE_NAME FROM information_schema.tables where "
                "TABLE_SCHEMA=database()"
            )
        )
        table_results = set(row[0] for row in table_results)  # noqa: C401
        self._all_tables = table_results
        self._metadata.reflect(bind=self._engine)
        return self._all_tables

    def get_grants(self):
        """Get grants."""
        cursor = self.get_session().execute(text("SHOW GRANTS"))
        grants = cursor.fetchall()
        if len(grants) == 0:
            return []
        if len(grants[0]) == 2:
            grants_list = [x[1] for x in grants]
        else:
            grants_list = [x[2] for x in grants]
        return grants_list

    def _get_current_version(self):
        """Get database current version."""
        return int(
            self.get_session().execute(text("select current_version()")).scalar()
        )

    def get_collation(self):
        """Get collation.

        ref `SHOW COLLATION <https://doris.apache.org/zh-CN/docs/dev/sql-manual/
        sql-reference/Show-Statements/SHOW-COLLATION/>`_

        """
        cursor = self.get_session().execute(text("SHOW COLLATION"))
        results = cursor.fetchall()
        return "" if not results else results[0][0]

    def get_users(self):
        """Get user info."""
        return []

    def get_columns(self, table_name: str) -> List[Dict]:
        """Get columns.

        Args:
            table_name (str): str
        Returns:
            columns: List[Dict], which contains name: str, type: str,
                default_expression: str, is_in_primary_key: bool, comment: str
                eg:[{'name': 'id', 'type': 'UInt64', 'default_expression': '',
                'is_in_primary_key': True, 'comment': 'id'}, ...]
        """
        fields = self.get_fields(table_name)
        return [
            {
                "name": field[0],
                "type": field[1],
                "default": field[2],
                "nullable": field[3],
                "comment": field[4],
            }
            for field in fields
        ]

    def get_fields(self, table_name) -> List[Tuple]:
        """Get column fields about specified table."""
        cursor = self.get_session().execute(
            text(
                "select COLUMN_NAME, COLUMN_TYPE, COLUMN_DEFAULT, IS_NULLABLE, "
                "COLUMN_COMMENT from information_schema.columns "
                f'where TABLE_NAME="{table_name}" and TABLE_SCHEMA=database()'
            )
        )
        fields = cursor.fetchall()
        return [(field[0], field[1], field[2], field[3], field[4]) for field in fields]

    def get_charset(self):
        """Get character_set."""
        return "utf-8"

    def get_show_create_table(self, table_name) -> str:
        """Get show create table."""
        # cur = self.get_session().execute(
        #     text(
        #         f"""show create table {table_name}"""
        #     )
        # )
        # rows = cur.fetchone()
        # create_sql = rows[1]
        # return create_sql
        # 这里是要表描述, 返回建表语句会导致token过长而失败
        cur = self.get_session().execute(
            text(
                f"SELECT TABLE_COMMENT "
                f"FROM information_schema.tables "
                f'where TABLE_NAME="{table_name}" and TABLE_SCHEMA=database()'
            )
        )
        table = cur.fetchone()
        if table:
            return str(table[0])
        else:
            return ""

    def get_table_comments(self, db_name=None):
        """Get table comments."""
        db_name = "database()" if not db_name else f"'{db_name}'"
        cursor = self.get_session().execute(
            text(
                f"SELECT TABLE_NAME,TABLE_COMMENT "
                f"FROM information_schema.tables "
                f"where TABLE_SCHEMA={db_name}"
            )
        )
        tables = cursor.fetchall()
        return [(table[0], table[1]) for table in tables]

    def get_database_names(self):
        """Get database names."""
        cursor = self.get_session().execute(text("SHOW DATABASES"))
        results = cursor.fetchall()
        return [
            d[0]
            for d in results
            if d[0]
            not in [
                "information_schema",
                "sys",
                "_statistics_",
                "mysql",
                "__internal_schema",
                "doris_audit_db__",
            ]
        ]

    def get_current_db_name(self) -> str:
        """Get current database name."""
        return self.get_session().execute(text("select database()")).scalar()

    def table_simple_info(self):
        """Get table simple info."""
        cursor = self.get_session().execute(
            text(
                "SELECT concat(TABLE_NAME,'(',group_concat(COLUMN_NAME,','),');') "
                "FROM information_schema.columns "
                "where TABLE_SCHEMA=database() "
                "GROUP BY TABLE_NAME"
            )
        )
        results = cursor.fetchall()
        return [x[0] for x in results]

    def get_indexes(self, table_name):
        """Get table indexes about specified table."""
        cursor = self.get_session().execute(text(f"SHOW INDEX FROM {table_name}"))
        indexes = cursor.fetchall()
        return [(index[2], index[4]) for index in indexes]



"""Role class for role-based conversation."""

from abc import ABC
from typing import Any, Dict, List, Optional, Set

from jinja2.meta import find_undeclared_variables
from jinja2.sandbox import SandboxedEnvironment

from dbgpt._private.pydantic import BaseModel, ConfigDict, Field

from .action.base import ActionOutput
from .memory.agent_memory import AgentMemory, AgentMemoryFragment
from .memory.llm import LLMImportanceScorer, LLMInsightExtractor
from .profile import Profile, ProfileConfig


class Role(ABC, BaseModel):
    """Role class for role-based conversation."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    profile: ProfileConfig = Field(
        ...,
        description="The profile of the role.",
    )
    memory: AgentMemory = Field(default_factory=AgentMemory)

    fixed_subgoal: Optional[str] = Field(None, description="Fixed subgoal")

    language: str = "en"
    is_human: bool = False
    is_team: bool = False

    template_env: SandboxedEnvironment = Field(default_factory=SandboxedEnvironment)

    async def build_prompt(
        self,
        question: Optional[str] = None,
        is_system: bool = True,
        most_recent_memories: Optional[str] = None,
        **kwargs
    ) -> str:
        """Return the prompt template for the role.

        Returns:
            str: The prompt template.
        """
        prompt_template = (
            self.system_prompt_template if is_system else self.user_prompt_template
        )
        template_vars = self._get_template_variables(prompt_template)
        _sub_render_keys = {"role", "name", "goal", "expand_prompt", "constraints"}
        pass_vars = {
            "role": self.role,
            "name": self.name,
            "goal": self.goal,
            "expand_prompt": self.expand_prompt,
            "language": self.language,
            "constraints": self.constraints,
            "most_recent_memories": (
                most_recent_memories if most_recent_memories else None
            ),
            "examples": self.examples,
            # "out_schema": out_schema if out_schema else None,
            # "resource_prompt": resource_prompt if resource_prompt else None,
            "question": question,
        }
        resource_vars = await self.generate_resource_variables(question)
        pass_vars.update(resource_vars)
        pass_vars.update(kwargs)
        filtered_data = {
            key: pass_vars[key] for key in template_vars if key in pass_vars
        }
        for key in filtered_data.keys():
            value = filtered_data[key]
            if key in _sub_render_keys and value:
                if isinstance(value, str):
                    # Render the sub-template
                    filtered_data[key] = self._render_template(value, **pass_vars)
                elif isinstance(value, list):
                    for i, item in enumerate(value):
                        if isinstance(item, str):
                            value[i] = self._render_template(item, **pass_vars)
        return self._render_template(prompt_template, **filtered_data)

    async def generate_resource_variables(
        self, question: Optional[str] = None
    ) -> Dict[str, Any]:
        """Generate the resource variables."""
        return {}

    def identity_check(self) -> None:
        """Check the identity of the role."""
        pass

    def get_name(self) -> str:
        """Get the name of the role."""
        return self.current_profile.get_name()

    @property
    def current_profile(self) -> Profile:
        """Return the current profile."""
        profile = self.profile.create_profile()
        return profile

    @property
    def name(self) -> str:
        """Return the name of the role."""
        return self.current_profile.get_name()

    @property
    def examples(self) -> Optional[str]:
        """Return the examples of the role."""
        return self.current_profile.get_examples()

    @property
    def role(self) -> str:
        """Return the role of the role."""
        return self.current_profile.get_role()

    @property
    def goal(self) -> Optional[str]:
        """Return the goal of the role."""
        return self.current_profile.get_goal()

    @property
    def constraints(self) -> Optional[List[str]]:
        """Return the constraints of the role."""
        return self.current_profile.get_constraints()

    @property
    def desc(self) -> Optional[str]:
        """Return the description of the role."""
        return self.current_profile.get_description()

    @property
    def expand_prompt(self) -> Optional[str]:
        """Return the expand prompt of the role."""
        return self.current_profile.get_expand_prompt()

    @property
    def system_prompt_template(self) -> str:
        """Return the current system prompt template."""
        return self.current_profile.get_system_prompt_template()

    @property
    def user_prompt_template(self) -> str:
        """Return the current user prompt template."""
        return self.current_profile.get_user_prompt_template()

    @property
    def save_memory_template(self) -> str:
        """Return the current save memory template."""
        return self.current_profile.get_save_memory_template()

    def _get_template_variables(self, template: str) -> Set[str]:
        parsed_content = self.template_env.parse(template)
        return find_undeclared_variables(parsed_content)

    def _render_template(self, template: str, **kwargs):
        r_template = self.template_env.from_string(template)
        return r_template.render(**kwargs)

    @property
    def memory_importance_scorer(self) -> Optional[LLMImportanceScorer]:
        """Create the memory importance scorer.

        The memory importance scorer is used to score the importance of a memory
        fragment.
        """
        return None

    @property
    def memory_insight_extractor(self) -> Optional[LLMInsightExtractor]:
        """Create the memory insight extractor.

        The memory insight extractor is used to extract a high-level insight from a
        memory fragment.
        """
        return None

    async def read_memories(
        self,
        question: str,
    ) -> str:
        """Read the memories from the memory."""
        memories = await self.memory.read(question)
        recent_messages = [m.raw_observation for m in memories]
        return "".join(recent_messages)

    async def save_to_memory(
        self,
        question: str,
        ai_message: str,
        action_output: Optional[ActionOutput] = None,
        check_pass: bool = True,
        check_fail_reason: Optional[str] = None,
    ) -> None:
        """Save the role to the memory."""
        if not action_output:
            raise ValueError("Action output is required to save to memory.")

        mem_thoughts = action_output.thoughts or ai_message
        observation = action_output.observations or action_output.content
        if not check_pass and check_fail_reason:
            observation += "\n" + check_fail_reason

        memory_map = {
            "question": question,
            "thought": mem_thoughts,
            "action": action_output.action,
            "observation": observation,
        }
        save_memory_template = self.save_memory_template
        memory_content = self._render_template(save_memory_template, **memory_map)
        fragment = AgentMemoryFragment(memory_content)
        await self.memory.write(fragment)



"""Agent memory module."""

from datetime import datetime
from typing import Callable, List, Optional, Type, cast

from dbgpt.core import LLMClient
from dbgpt.util.annotations import immutable, mutable
from dbgpt.util.id_generator import new_id

from .base import (
    DiscardedMemoryFragments,
    ImportanceScorer,
    InsightExtractor,
    Memory,
    MemoryFragment,
    ShortTermMemory,
    WriteOperation,
)
from .gpts import GptsMemory, GptsMessageMemory, GptsPlansMemory


class AgentMemoryFragment(MemoryFragment):
    """Default memory fragment for agent memory."""

    def __init__(
        self,
        observation: str,
        embeddings: Optional[List[float]] = None,
        memory_id: Optional[int] = None,
        importance: Optional[float] = None,
        last_accessed_time: Optional[datetime] = None,
        is_insight: bool = False,
    ):
        """Create a memory fragment."""
        if not memory_id:
            # Generate a new memory id, we use snowflake id generator here.
            memory_id = new_id()
        self.observation = observation
        self._embeddings = embeddings
        self.memory_id: int = cast(int, memory_id)
        self._importance: Optional[float] = importance
        self._last_accessed_time: Optional[datetime] = last_accessed_time
        self._is_insight = is_insight

    @property
    def id(self) -> int:
        """Return the memory id."""
        return self.memory_id

    @property
    def raw_observation(self) -> str:
        """Return the raw observation."""
        return self.observation

    @property
    def embeddings(self) -> Optional[List[float]]:
        """Return the embeddings of the memory fragment."""
        return self._embeddings

    def update_embeddings(self, embeddings: List[float]) -> None:
        """Update the embeddings of the memory fragment.

        Args:
            embeddings(List[float]): embeddings
        """
        self._embeddings = embeddings

    def calculate_current_embeddings(
        self, embedding_func: Callable[[List[str]], List[List[float]]]
    ) -> List[float]:
        """Calculate the embeddings of the memory fragment.

        Args:
            embedding_func(Callable[[List[str]], List[List[float]]]): Function to
                compute embeddings

        Returns:
            List[float]: Embeddings of the memory fragment
        """
        embeddings = embedding_func([self.observation])
        return embeddings[0]

    @property
    def is_insight(self) -> bool:
        """Return whether the memory fragment is an insight.

        Returns:
            bool: Whether the memory fragment is an insight
        """
        return self._is_insight

    @property
    def importance(self) -> Optional[float]:
        """Return the importance of the memory fragment.

        Returns:
            Optional[float]: Importance of the memory fragment
        """
        return self._importance

    def update_importance(self, importance: float) -> Optional[float]:
        """Update the importance of the memory fragment.

        Args:
            importance(float): Importance of the memory fragment

        Returns:
            Optional[float]: Old importance
        """
        old_importance = self._importance
        self._importance = importance
        return old_importance

    @property
    def last_accessed_time(self) -> Optional[datetime]:
        """Return the last accessed time of the memory fragment.

        Used to determine the least recently used memory fragment.

        Returns:
            Optional[datetime]: Last accessed time
        """
        return self._last_accessed_time

    def update_accessed_time(self, now: datetime) -> Optional[datetime]:
        """Update the last accessed time of the memory fragment.

        Args:
            now(datetime): Current time

        Returns:
            Optional[datetime]: Old last accessed time
        """
        old_time = self._last_accessed_time
        self._last_accessed_time = now
        return old_time

    @classmethod
    def build_from(
        cls: Type["AgentMemoryFragment"],
        observation: str,
        embeddings: Optional[List[float]] = None,
        memory_id: Optional[int] = None,
        importance: Optional[float] = None,
        is_insight: bool = False,
        last_accessed_time: Optional[datetime] = None,
        **kwargs
    ) -> "AgentMemoryFragment":
        """Build a memory fragment from the given parameters."""
        return cls(
            observation=observation,
            embeddings=embeddings,
            memory_id=memory_id,
            importance=importance,
            last_accessed_time=last_accessed_time,
            is_insight=is_insight,
        )

    def copy(self: "AgentMemoryFragment") -> "AgentMemoryFragment":
        """Return a copy of the memory fragment."""
        return AgentMemoryFragment.build_from(
            observation=self.observation,
            embeddings=self._embeddings,
            memory_id=self.memory_id,
            importance=self.importance,
            last_accessed_time=self.last_accessed_time,
            is_insight=self.is_insight,
        )


class AgentMemory(Memory[AgentMemoryFragment]):
    """Agent memory."""

    def __init__(
        self,
        memory: Optional[Memory[AgentMemoryFragment]] = None,
        importance_scorer: Optional[ImportanceScorer[AgentMemoryFragment]] = None,
        insight_extractor: Optional[InsightExtractor[AgentMemoryFragment]] = None,
        gpts_memory: Optional[GptsMemory] = None,
    ):
        """Create an agent memory.

        Args:
            memory(Memory[AgentMemoryFragment]): Memory to store fragments
            importance_scorer(ImportanceScorer[AgentMemoryFragment]): Scorer to
                calculate the importance of memory fragments
            insight_extractor(InsightExtractor[AgentMemoryFragment]): Extractor to
                extract insights from memory fragments
            gpts_memory(GptsMemory): Memory to store GPTs related information
        """
        if not memory:
            memory = ShortTermMemory(buffer_size=5)
        if not gpts_memory:
            gpts_memory = GptsMemory()
        self.memory: Memory[AgentMemoryFragment] = cast(
            Memory[AgentMemoryFragment], memory
        )
        self.importance_scorer = importance_scorer
        self.insight_extractor = insight_extractor
        self.gpts_memory = gpts_memory

    @immutable
    def structure_clone(
        self: "AgentMemory", now: Optional[datetime] = None
    ) -> "AgentMemory":
        """Return a structure clone of the memory.

        The gpst_memory is not cloned, it will be shared in whole agent memory.
        """
        m = AgentMemory(
            memory=self.memory.structure_clone(now),
            importance_scorer=self.importance_scorer,
            insight_extractor=self.insight_extractor,
            gpts_memory=self.gpts_memory,
        )
        m._copy_from(self)
        return m

    @mutable
    def initialize(
        self,
        name: Optional[str] = None,
        llm_client: Optional[LLMClient] = None,
        importance_scorer: Optional[ImportanceScorer[AgentMemoryFragment]] = None,
        insight_extractor: Optional[InsightExtractor[AgentMemoryFragment]] = None,
        real_memory_fragment_class: Optional[Type[AgentMemoryFragment]] = None,
    ) -> None:
        """Initialize the memory."""
        self.memory.initialize(
            name=name,
            llm_client=llm_client,
            importance_scorer=importance_scorer or self.importance_scorer,
            insight_extractor=insight_extractor or self.insight_extractor,
            real_memory_fragment_class=real_memory_fragment_class
            or AgentMemoryFragment,
        )

    @mutable
    async def write(
        self,
        memory_fragment: AgentMemoryFragment,
        now: Optional[datetime] = None,
        op: WriteOperation = WriteOperation.ADD,
    ) -> Optional[DiscardedMemoryFragments[AgentMemoryFragment]]:
        """Write a memory fragment to the memory."""
        return await self.memory.write(memory_fragment, now)

    @immutable
    async def read(
        self,
        observation: str,
        alpha: Optional[float] = None,
        beta: Optional[float] = None,
        gamma: Optional[float] = None,
    ) -> List[AgentMemoryFragment]:
        """Read memory fragments related to the observation.

        Args:
            observation(str): Observation
            alpha(float): Importance weight
            beta(float): Time weight
            gamma(float): Randomness weight

        Returns:
            List[AgentMemoryFragment]: List of memory fragments
        """
        return await self.memory.read(observation, alpha, beta, gamma)

    @mutable
    async def clear(self) -> List[AgentMemoryFragment]:
        """Clear the memory."""
        return await self.memory.clear()

    @property
    def plans_memory(self) -> GptsPlansMemory:
        """Return the plan memory."""
        return self.gpts_memory.plans_memory

    @property
    def message_memory(self) -> GptsMessageMemory:
        """Return the message memory."""
        return self.gpts_memory.message_memory



"""Host tool resource module."""

from ...resource.tool.base import tool


@tool(description="Get current host CPU status.")
def get_current_host_cpu_status() -> str:
    """Get current host CPU status."""
    import platform

    import psutil

    cpu_architecture = platform.machine()
    cpu_count_physical = psutil.cpu_count(logical=False)
    cpu_count_logical = psutil.cpu_count(logical=True)
    cpu_usage = psutil.cpu_percent(interval=1)
    return (
        f"CPU Architecture: {cpu_architecture}\n"
        f"Physical CPU Cores: {cpu_count_physical}\n"
        f"Logical CPU Cores: {cpu_count_logical}\n"
        f"CPU Usage: {cpu_usage}%"
    )


@tool(description="Get current host memory status.")
def get_current_host_memory_status() -> str:
    """Get current host memory status."""
    import psutil

    memory = psutil.virtual_memory()
    return (
        f"Total:  {memory.total / (1024**3):.2f} GB\n"
        f"Available: {memory.available / (1024**3):.2f} GB\n"
        f"Used:  {memory.used / (1024**3):.2f} GB\n"
        f"Percent: {memory.percent}%"
    )


@tool(description="Get current host system load.")
def get_current_host_system_load() -> str:
    """Get current host system load."""
    import os

    load1, load5, load15 = os.getloadavg()
    return f"System load average: {load1:.2f}, {load5:.2f}, {load15:.2f}"



"""Base class for resource group."""
from abc import ABC, abstractmethod


class ResourceGroup(ABC):
    """Base class for resource group.

    A resource group is a group of resources that are related to each other.
    It contains the all resources that are needed to run a workflow.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Return the name of current resource group."""



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List, Optional, Union

if TYPE_CHECKING:
    from dbgpt.core.interface.message import BaseMessage, ModelMessage

logger = logging.getLogger(__name__)


class ProxyTokenizerWrapper:
    def __init__(self) -> None:
        self._support_encoding = True
        self._encoding_model = None

    def count_token(
        self,
        messages: Union[str, BaseMessage, ModelMessage, List[ModelMessage]],
        model_name: Optional[str] = None,
    ) -> int:
        """Count token of given messages

        Args:
            messages (Union[str, BaseMessage, ModelMessage, List[ModelMessage]]): messages to count token
            model_name (Optional[str], optional): model name. Defaults to None.

        Returns:
            int: token count, -1 if failed
        """
        if not self._support_encoding:
            logger.warning(
                "model does not support encoding model, can't count token, returning -1"
            )
            return -1
        encoding = self._get_or_create_encoding_model(model_name)
        cnt = 0
        if isinstance(messages, str):
            cnt = len(encoding.encode(messages, disallowed_special=()))
        elif isinstance(messages, BaseMessage):
            cnt = len(encoding.encode(messages.content, disallowed_special=()))
        elif isinstance(messages, ModelMessage):
            cnt = len(encoding.encode(messages.content, disallowed_special=()))
        elif isinstance(messages, list):
            for message in messages:
                cnt += len(encoding.encode(message.content, disallowed_special=()))
        else:
            logger.warning(
                "unsupported type of messages, can't count token, returning -1"
            )
            return -1
        return cnt

    def _get_or_create_encoding_model(self, model_name: Optional[str] = None):
        """Get or create encoding model for given model name
        More detail see: https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
        """
        if self._encoding_model:
            return self._encoding_model
        try:
            import tiktoken

            logger.info(
                "tiktoken installed, using it to count tokens, tiktoken will download tokenizer from network, "
                "also you can download it and put it in the directory of environment variable TIKTOKEN_CACHE_DIR"
            )
        except ImportError:
            self._support_encoding = False
            logger.warn("tiktoken not installed, cannot count tokens, returning -1")
            return -1
        try:
            if not model_name:
                model_name = "gpt-3.5-turbo"
            self._encoding_model = tiktoken.model.encoding_for_model(model_name)
        except KeyError:
            logger.warning(
                f"{model_name}'s tokenizer not found, using cl100k_base encoding."
            )
            self._encoding_model = tiktoken.get_encoding("cl100k_base")
        return self._encoding_model



"""Datasource Knowledge."""
from typing import Any, Dict, List, Optional, Union

from dbgpt.core import Document
from dbgpt.datasource import BaseConnector

from ..summary.gdbms_db_summary import _parse_db_summary as _parse_gdb_summary
from ..summary.rdbms_db_summary import _parse_db_summary
from .base import ChunkStrategy, DocumentType, Knowledge, KnowledgeType


class DatasourceKnowledge(Knowledge):
    """Datasource Knowledge."""

    def __init__(
        self,
        connector: BaseConnector,
        summary_template: str = "{table_name}({columns})",
        knowledge_type: Optional[KnowledgeType] = KnowledgeType.DOCUMENT,
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
        **kwargs: Any,
    ) -> None:
        """Create Datasource Knowledge with Knowledge arguments.

        Args:
            connector(BaseConnector): connector
            summary_template(str, optional): summary template
            knowledge_type(KnowledgeType, optional): knowledge type
            metadata(Dict[str, Union[str, List[str]], optional): metadata
        """
        self._connector = connector
        self._summary_template = summary_template
        super().__init__(knowledge_type=knowledge_type, metadata=metadata, **kwargs)

    def _load(self) -> List[Document]:
        """Load datasource document from data_loader."""
        docs = []
        if self._connector.is_graph_type():
            db_summary = _parse_gdb_summary(self._connector, self._summary_template)
        else:
            db_summary = _parse_db_summary(self._connector, self._summary_template)
        for table_summary in db_summary:
            metadata = {"source": "database"}
            if self._metadata:
                metadata.update(self._metadata)  # type: ignore
            docs.append(Document(content=table_summary, metadata=metadata))
        return docs

    @classmethod
    def support_chunk_strategy(cls) -> List[ChunkStrategy]:
        """Return support chunk strategy."""
        return [
            ChunkStrategy.CHUNK_BY_SIZE,
            ChunkStrategy.CHUNK_BY_SEPARATOR,
        ]

    @classmethod
    def type(cls) -> KnowledgeType:
        """Knowledge type of Datasource."""
        return KnowledgeType.DOCUMENT

    @classmethod
    def document_type(cls) -> DocumentType:
        """Return document type."""
        return DocumentType.DATASOURCE



"""Summary classes for database, table, field and index."""
from typing import Dict, Iterable, List, Optional, Tuple


class DBSummary:
    """Database summary class."""

    def __init__(self, name: str):
        """Create a new DBSummary."""
        self.name = name
        self.summary: Optional[str] = None
        self.tables: Iterable[str] = []
        self.metadata: Optional[str] = None

    def get_summary(self) -> Optional[str]:
        """Get the summary."""
        return self.summary


class TableSummary:
    """Table summary class."""

    def __init__(self, name: str):
        """Create a new TableSummary."""
        self.name = name
        self.summary: Optional[str] = None
        self.fields: List[Tuple] = []
        self.indexes: List[Dict] = []


class FieldSummary:
    """Field summary class."""

    def __init__(self, name: str):
        """Create a new FieldSummary."""
        self.name = name
        self.summary = None
        self.data_type = None


class IndexSummary:
    """Index summary class."""

    def __init__(self, name: str):
        """Create a new IndexSummary."""
        self.name = name
        self.summary: Optional[str] = None
        self.bind_fields: List[str] = []



"""Chat history database model."""
from datetime import datetime
from typing import Optional

from sqlalchemy import Column, DateTime, Index, Integer, String, Text, UniqueConstraint

from ..metadata import BaseDao, Model


class ChatHistoryEntity(Model):
    """Chat history entity."""

    __tablename__ = "chat_history"
    __table_args__ = (UniqueConstraint("conv_uid", name="uk_conv_uid"),)
    id = Column(
        Integer, primary_key=True, autoincrement=True, comment="autoincrement id"
    )
    conv_uid = Column(
        String(255),
        # Change from False to True, the alembic migration will fail, so we use
        # UniqueConstraint to replace it
        unique=False,
        nullable=False,
        comment="Conversation record unique id",
    )
    chat_mode = Column(String(255), nullable=False, comment="Conversation scene mode")
    summary = Column(
        Text(length=2**31 - 1), nullable=False, comment="Conversation record summary"
    )
    user_name = Column(String(255), nullable=True, comment="interlocutor")
    messages = Column(
        Text(length=2**31 - 1), nullable=True, comment="Conversation details"
    )
    message_ids = Column(
        Text(length=2**31 - 1), nullable=True, comment="Message ids, split by comma"
    )
    sys_code = Column(String(128), index=True, nullable=True, comment="System code")
    gmt_created = Column(DateTime, default=datetime.now, comment="Record creation time")
    gmt_modified = Column(DateTime, default=datetime.now, comment="Record update time")

    Index("idx_q_user", "user_name")
    Index("idx_q_mode", "chat_mode")
    Index("idx_q_conv", "summary")


class ChatHistoryMessageEntity(Model):
    """Chat history message entity."""

    __tablename__ = "chat_history_message"
    __table_args__ = (
        UniqueConstraint("conv_uid", "index", name="uk_conversation_message"),
    )
    id = Column(
        Integer, primary_key=True, autoincrement=True, comment="autoincrement id"
    )
    conv_uid = Column(
        String(255),
        unique=False,
        nullable=False,
        comment="Conversation record unique id",
    )
    index = Column(Integer, nullable=False, comment="Message index")
    round_index = Column(Integer, nullable=False, comment="Message round index")
    message_detail = Column(
        Text(length=2**31 - 1), nullable=True, comment="Message details, json format"
    )
    gmt_created = Column(DateTime, default=datetime.now, comment="Record creation time")
    gmt_modified = Column(DateTime, default=datetime.now, comment="Record update time")


class ChatHistoryDao(BaseDao):
    """Chat history dao."""

    def list_last_20(
        self, user_name: Optional[str] = None, sys_code: Optional[str] = None
    ):
        """Retrieve the last 20 chat history records."""
        session = self.get_raw_session()
        chat_history = session.query(ChatHistoryEntity)
        if user_name:
            chat_history = chat_history.filter(ChatHistoryEntity.user_name == user_name)
        if sys_code:
            chat_history = chat_history.filter(ChatHistoryEntity.sys_code == sys_code)

        chat_history = chat_history.order_by(ChatHistoryEntity.id.desc())

        result = chat_history.limit(20).all()
        session.close()
        return result

    def raw_update(self, entity: ChatHistoryEntity):
        """Update the chat history record."""
        session = self.get_raw_session()
        try:
            updated = session.merge(entity)
            session.commit()
            return updated.id
        finally:
            session.close()

    def update_message_by_uid(self, message: str, conv_uid: str):
        """Update the chat history record."""
        session = self.get_raw_session()
        try:
            chat_history = session.query(ChatHistoryEntity)
            chat_history = chat_history.filter(ChatHistoryEntity.conv_uid == conv_uid)
            updated = chat_history.update({ChatHistoryEntity.messages: message})
            session.commit()
            return updated
        finally:
            session.close()

    def raw_delete(self, conv_uid: str):
        """Delete the chat history record."""
        if conv_uid is None:
            raise Exception("conv_uid is None")
        with self.session() as session:
            chat_history = session.query(ChatHistoryEntity)
            chat_history = chat_history.filter(ChatHistoryEntity.conv_uid == conv_uid)
            chat_history.delete()

    def get_by_uid(self, conv_uid: str) -> Optional[ChatHistoryEntity]:
        """Retrieve the chat history record by conv_uid."""
        with self.session(commit=False) as session:
            return session.query(ChatHistoryEntity).filter_by(conv_uid=conv_uid).first()



from __future__ import annotations

import logging
from typing import List, Optional, Type, cast

from dbgpt.configs.model_config import get_device
from dbgpt.core import Embeddings, RerankEmbeddings
from dbgpt.model.parameter import (
    BaseEmbeddingModelParameters,
    EmbeddingModelParameters,
    ProxyEmbeddingParameters,
)
from dbgpt.util.parameter_utils import EnvArgumentParser, _get_dict_from_obj
from dbgpt.util.system_utils import get_system_info
from dbgpt.util.tracer import SpanType, SpanTypeRunName, root_tracer

logger = logging.getLogger(__name__)


class EmbeddingLoader:
    def __init__(self) -> None:
        pass

    def load(self, model_name: str, param: BaseEmbeddingModelParameters) -> Embeddings:
        metadata = {
            "model_name": model_name,
            "run_service": SpanTypeRunName.EMBEDDING_MODEL.value,
            "params": _get_dict_from_obj(param),
            "sys_infos": _get_dict_from_obj(get_system_info()),
        }
        with root_tracer.start_span(
            "EmbeddingLoader.load", span_type=SpanType.RUN, metadata=metadata
        ):
            # add more models
            if model_name in ["proxy_openai", "proxy_azure"]:
                from langchain.embeddings import OpenAIEmbeddings

                from dbgpt.rag.embedding._wrapped import WrappedEmbeddings

                return WrappedEmbeddings(OpenAIEmbeddings(**param.build_kwargs()))
            elif model_name in ["proxy_http_openapi"]:
                from dbgpt.rag.embedding import OpenAPIEmbeddings

                proxy_param = cast(ProxyEmbeddingParameters, param)
                openapi_param = {}
                if proxy_param.proxy_server_url:
                    openapi_param["api_url"] = proxy_param.proxy_server_url
                if proxy_param.proxy_api_key:
                    openapi_param["api_key"] = proxy_param.proxy_api_key
                if proxy_param.proxy_backend:
                    openapi_param["model_name"] = proxy_param.proxy_backend
                return OpenAPIEmbeddings(**openapi_param)
            elif model_name in ["proxy_ollama"]:
                from dbgpt.rag.embedding import OllamaEmbeddings

                proxy_param = cast(ProxyEmbeddingParameters, param)
                ollama_param = {}
                if proxy_param.proxy_server_url:
                    ollama_param["api_url"] = proxy_param.proxy_server_url
                if proxy_param.proxy_backend:
                    ollama_param["model_name"] = proxy_param.proxy_backend
                return OllamaEmbeddings(**ollama_param)
            else:
                from dbgpt.rag.embedding import HuggingFaceEmbeddings

                kwargs = param.build_kwargs(model_name=param.model_path)
                return HuggingFaceEmbeddings(**kwargs)

    def load_rerank_model(
        self, model_name: str, param: BaseEmbeddingModelParameters
    ) -> RerankEmbeddings:
        metadata = {
            "model_name": model_name,
            "run_service": SpanTypeRunName.EMBEDDING_MODEL.value,
            "params": _get_dict_from_obj(param),
            "sys_infos": _get_dict_from_obj(get_system_info()),
        }
        with root_tracer.start_span(
            "EmbeddingLoader.load_rerank_model",
            span_type=SpanType.RUN,
            metadata=metadata,
        ):
            if model_name in ["rerank_proxy_http_openapi"]:
                from dbgpt.rag.embedding.rerank import OpenAPIRerankEmbeddings

                proxy_param = cast(ProxyEmbeddingParameters, param)
                openapi_param = {}
                if proxy_param.proxy_server_url:
                    openapi_param["api_url"] = proxy_param.proxy_server_url
                if proxy_param.proxy_api_key:
                    openapi_param["api_key"] = proxy_param.proxy_api_key
                if proxy_param.proxy_backend:
                    openapi_param["model_name"] = proxy_param.proxy_backend
                return OpenAPIRerankEmbeddings(**openapi_param)
            else:
                from dbgpt.rag.embedding.rerank import CrossEncoderRerankEmbeddings

                kwargs = param.build_kwargs(model_name=param.model_path)
                return CrossEncoderRerankEmbeddings(**kwargs)


def _parse_embedding_params(
    model_name: Optional[str] = None,
    model_path: Optional[str] = None,
    command_args: List[str] = None,
    param_cls: Optional[Type] = EmbeddingModelParameters,
    **kwargs,
):
    model_args = EnvArgumentParser()
    env_prefix = EnvArgumentParser.get_env_prefix(model_name)
    model_params: BaseEmbeddingModelParameters = model_args.parse_args_into_dataclass(
        param_cls,
        env_prefixes=[env_prefix],
        command_args=command_args,
        model_name=model_name,
        model_path=model_path,
        **kwargs,
    )
    if not model_params.device:
        model_params.device = get_device()
        logger.info(
            f"[EmbeddingsModelWorker] Parameters of device is None, use {model_params.device}"
        )
    return model_params



"""Knowledge Factory to create knowledge from file path and url."""
from typing import Dict, List, Optional, Type, Union

from dbgpt.rag.knowledge.base import Knowledge, KnowledgeType
from dbgpt.rag.knowledge.string import StringKnowledge
from dbgpt.rag.knowledge.url import URLKnowledge


class KnowledgeFactory:
    """Knowledge Factory to create knowledge from file path and url."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        knowledge_type: Optional[KnowledgeType] = KnowledgeType.DOCUMENT,
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
    ):
        """Create Knowledge Factory with file path and knowledge type.

        Args:
            file_path(str, optional): file path
            knowledge_type(KnowledgeType, optional): knowledge type
        """
        self._file_path = file_path
        self._knowledge_type = knowledge_type
        self._metadata = metadata

    @classmethod
    def create(
        cls,
        datasource: str = "",
        knowledge_type: KnowledgeType = KnowledgeType.DOCUMENT,
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
    ):
        """Create knowledge from file path, url or text.

        Args:
             datasource: path of the file to convert
             knowledge_type: type of knowledge
             metadata: Optional[Dict[str, Union[str, List[str]]]]

        Examples:
            .. code-block:: python

                from dbgpt.rag.knowledge.factory import KnowledgeFactory

                url_knowlege = KnowledgeFactory.create(
                    datasource="https://www.baidu.com", knowledge_type=KnowledgeType.URL
                )
                doc_knowlege = KnowledgeFactory.create(
                    datasource="path/to/document.pdf",
                    knowledge_type=KnowledgeType.DOCUMENT,
                )

        """
        match knowledge_type:
            case KnowledgeType.DOCUMENT:
                return cls.from_file_path(
                    file_path=datasource,
                    knowledge_type=knowledge_type,
                    metadata=metadata,
                )
            case KnowledgeType.URL:
                return cls.from_url(url=datasource, knowledge_type=knowledge_type)
            case KnowledgeType.TEXT:
                return cls.from_text(
                    text=datasource, knowledge_type=knowledge_type, metadata=metadata
                )
            case _:
                raise Exception(f"Unsupported knowledge type '{knowledge_type}'")

    @classmethod
    def from_file_path(
        cls,
        file_path: str = "",
        knowledge_type: Optional[KnowledgeType] = KnowledgeType.DOCUMENT,
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
    ) -> Knowledge:
        """Create knowledge from path.

        Args:
            param file_path: path of the file to convert
            param knowledge_type: type of knowledge

        Examples:
            .. code-block:: python

                from dbgpt.rag.knowledge.factory import KnowledgeFactory

                doc_knowlege = KnowledgeFactory.create(
                    datasource="path/to/document.pdf",
                    knowledge_type=KnowledgeType.DOCUMENT,
                )

        """
        factory = cls(file_path=file_path, knowledge_type=knowledge_type)
        return factory._select_document_knowledge(
            file_path=file_path, knowledge_type=knowledge_type, metadata=metadata
        )

    @staticmethod
    def from_url(
        url: str = "",
        knowledge_type: KnowledgeType = KnowledgeType.URL,
    ) -> Knowledge:
        """Create knowledge from url.

        Args:
            param url: url of the file to convert
            param knowledge_type: type of knowledge

        Examples:
            .. code-block:: python

                from dbgpt.rag.knowledge.factory import KnowledgeFactory

                url_knowlege = KnowledgeFactory.create(
                    datasource="https://www.baidu.com", knowledge_type=KnowledgeType.URL
                )
        """
        return URLKnowledge(
            url=url,
            knowledge_type=knowledge_type,
        )

    @staticmethod
    def from_text(
        text: str = "",
        knowledge_type: KnowledgeType = KnowledgeType.TEXT,
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
    ) -> Knowledge:
        """Create knowledge from text.

        Args:
            param text: text to convert
            param knowledge_type: type of knowledge
        """
        return StringKnowledge(
            text=text,
            knowledge_type=knowledge_type,
            metadata=metadata,
        )

    def _select_document_knowledge(self, **kwargs):
        """Select document knowledge from file path."""
        extension = self._file_path.rsplit(".", 1)[-1]
        knowledge_classes = self._get_knowledge_subclasses()
        implementation = None
        for cls in knowledge_classes:
            if cls.document_type() and cls.document_type().value == extension:
                implementation = cls(**kwargs)
        if implementation is None:
            raise Exception(f"Unsupported knowledge document type '{extension}'")
        return implementation

    @classmethod
    def all_types(cls):
        """Get all knowledge types."""
        return [knowledge.type().value for knowledge in cls._get_knowledge_subclasses()]

    @classmethod
    def subclasses(cls) -> List["Type[Knowledge]"]:
        """Get all knowledge subclasses."""
        return cls._get_knowledge_subclasses()

    @staticmethod
    def _get_knowledge_subclasses() -> List["Type[Knowledge]"]:
        """Get all knowledge subclasses."""
        from dbgpt.rag.knowledge.base import Knowledge  # noqa: F401
        from dbgpt.rag.knowledge.csv import CSVKnowledge  # noqa: F401
        from dbgpt.rag.knowledge.datasource import DatasourceKnowledge  # noqa: F401
        from dbgpt.rag.knowledge.docx import DocxKnowledge  # noqa: F401
        from dbgpt.rag.knowledge.excel import ExcelKnowledge  # noqa: F401
        from dbgpt.rag.knowledge.html import HTMLKnowledge  # noqa: F401
        from dbgpt.rag.knowledge.markdown import MarkdownKnowledge  # noqa: F401
        from dbgpt.rag.knowledge.pdf import PDFKnowledge  # noqa: F401
        from dbgpt.rag.knowledge.pptx import PPTXKnowledge  # noqa: F401
        from dbgpt.rag.knowledge.string import StringKnowledge  # noqa: F401
        from dbgpt.rag.knowledge.txt import TXTKnowledge  # noqa: F401
        from dbgpt.rag.knowledge.url import URLKnowledge  # noqa: F401

        return Knowledge.__subclasses__()



"""OceanBase vector store."""
import json
import logging
import os
import threading
import uuid
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple

from pydantic import Field
from sqlalchemy import Column, Table, create_engine, insert, text
from sqlalchemy.dialects.mysql import JSON, LONGTEXT, VARCHAR
from sqlalchemy.types import String, UserDefinedType

from dbgpt.core import Chunk, Document, Embeddings
from dbgpt.core.awel.flow import Parameter, ResourceCategory, register_resource
from dbgpt.storage.vector_store.base import (
    _COMMON_PARAMETERS,
    VectorStoreBase,
    VectorStoreConfig,
)
from dbgpt.storage.vector_store.filters import MetadataFilters
from dbgpt.util.i18n_utils import _

try:
    from sqlalchemy.orm import declarative_base
except ImportError:
    from sqlalchemy.ext.declarative import declarative_base


logger = logging.getLogger(__name__)
sql_logger = None
sql_dbg_log_path = os.getenv("OB_SQL_DBG_LOG_PATH", "")
if sql_dbg_log_path != "":
    sql_logger = logging.getLogger("ob_sql_dbg")
    sql_logger.setLevel(logging.DEBUG)
    file_handler = logging.FileHandler(sql_dbg_log_path)
    file_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setFormatter(formatter)
    sql_logger.addHandler(file_handler)

_OCEANBASE_DEFAULT_EMBEDDING_DIM = 1536
_OCEANBASE_DEFAULT_COLLECTION_NAME = "langchain_document"
_OCEANBASE_DEFAULT_IVFFLAT_ROW_THRESHOLD = 10000
_OCEANBASE_DEFAULT_RWLOCK_MAX_READER = 64

Base = declarative_base()


def ob_vector_from_db(value):
    """Parse vector from oceanbase."""
    return [float(v) for v in value[1:-1].split(",")]


def ob_vector_to_db(value, dim=None):
    """Parse vector to oceanbase vector constant type."""
    if value is None:
        return value

    return "[" + ",".join([str(float(v)) for v in value]) + "]"


class Vector(UserDefinedType):
    """OceanBase Vector Column Type."""

    cache_ok = True
    _string = String()

    def __init__(self, dim):
        """Create a Vector column with dimemsion `dim`."""
        super(UserDefinedType, self).__init__()
        self.dim = dim

    def get_col_spec(self, **kw):
        """Get vector column definition in string format."""
        return "VECTOR(%d)" % self.dim

    def bind_processor(self, dialect):
        """Get a processor to parse an array to oceanbase vector."""

        def process(value):
            return ob_vector_to_db(value, self.dim)

        return process

    def literal_processor(self, dialect):
        """Get a string processor to parse an array to OceanBase Vector."""
        string_literal_processor = self._string._cached_literal_processor(dialect)

        def process(value):
            return string_literal_processor(ob_vector_to_db(value, self.dim))

        return process

    def result_processor(self, dialect, coltype):
        """Get a processor to parse OceanBase Vector to array."""

        def process(value):
            return ob_vector_from_db(value)

        return process


class OceanBaseCollectionStat:
    """A tracer that maintains a table status in OceanBase."""

    def __init__(self):
        """Create OceanBaseCollectionStat instance."""
        self._lock = threading.Lock()
        self.maybe_collection_not_exist = True
        self.maybe_collection_index_not_exist = True

    def collection_exists(self):
        """Set a table is existing."""
        with self._lock:
            self.maybe_collection_not_exist = False

    def collection_index_exists(self):
        """Set the index of a table is existing."""
        with self._lock:
            self.maybe_collection_index_not_exist = False

    def collection_not_exists(self):
        """Set a table is dropped."""
        with self._lock:
            self.maybe_collection_not_exist = True

    def collection_index_not_exists(self):
        """Set the index of a table is dropped."""
        with self._lock:
            self.maybe_collection_index_not_exist = True

    def get_maybe_collection_not_exist(self):
        """Get the existing status of a table."""
        with self._lock:
            return self.maybe_collection_not_exist

    def get_maybe_collection_index_not_exist(self):
        """Get the existing stats of the index of a table."""
        with self._lock:
            return self.maybe_collection_index_not_exist


class OceanBaseGlobalRWLock:
    """A global rwlock for OceanBase to do creating vector index table offline ddl."""

    def __init__(self, max_readers) -> None:
        """Create a rwlock."""
        self.max_readers_ = max_readers
        self.writer_entered_ = False
        self.reader_cnt_ = 0
        self.mutex_ = threading.Lock()
        self.writer_cv_ = threading.Condition(self.mutex_)
        self.reader_cv_ = threading.Condition(self.mutex_)

    def rlock(self):
        """Lock for reading."""
        self.mutex_.acquire()
        while self.writer_entered_ or self.max_readers_ == self.reader_cnt_:
            self.reader_cv_.wait()
        self.reader_cnt_ += 1
        self.mutex_.release()

    def runlock(self):
        """Unlock reading lock."""
        self.mutex_.acquire()
        self.reader_cnt_ -= 1
        if self.writer_entered_:
            if self.reader_cnt_ == 0:
                self.writer_cv_.notify(1)
        else:
            if self.max_readers_ - 1 == self.reader_cnt_:
                self.reader_cv_.notify(1)
        self.mutex_.release()

    def wlock(self):
        """Lock for writing."""
        self.mutex_.acquire()
        while self.writer_entered_:
            self.reader_cv_.wait()
        self.writer_entered_ = True
        while 0 < self.reader_cnt_:
            self.writer_cv_.wait()
        self.mutex_.release()

    def wunlock(self):
        """Unlock writing lock."""
        self.mutex_.acquire()
        self.writer_entered_ = False
        self.reader_cv_.notifyAll()
        self.mutex_.release()

    class OBRLock:
        """Reading Lock Wrapper for `with` clause."""

        def __init__(self, rwlock) -> None:
            """Create reading lock wrapper instance."""
            self.rwlock_ = rwlock

        def __enter__(self):
            """Lock."""
            self.rwlock_.rlock()

        def __exit__(self, exc_type, exc_value, traceback):
            """Unlock when exiting."""
            self.rwlock_.runlock()

    class OBWLock:
        """Writing Lock Wrapper for `with` clause."""

        def __init__(self, rwlock) -> None:
            """Create writing lock wrapper instance."""
            self.rwlock_ = rwlock

        def __enter__(self):
            """Lock."""
            self.rwlock_.wlock()

        def __exit__(self, exc_type, exc_value, traceback):
            """Unlock when exiting."""
            self.rwlock_.wunlock()

    def reader_lock(self):
        """Get reading lock wrapper."""
        return self.OBRLock(self)

    def writer_lock(self):
        """Get writing lock wrapper."""
        return self.OBWLock(self)


ob_grwlock = OceanBaseGlobalRWLock(_OCEANBASE_DEFAULT_RWLOCK_MAX_READER)


class OceanBase:
    """OceanBase Vector Store implementation."""

    def __init__(
        self,
        database: str,
        connection_string: str,
        embedding_function: Embeddings,
        embedding_dimension: int = _OCEANBASE_DEFAULT_EMBEDDING_DIM,
        collection_name: str = _OCEANBASE_DEFAULT_COLLECTION_NAME,
        pre_delete_collection: bool = False,
        logger: Optional[logging.Logger] = None,
        engine_args: Optional[dict] = None,
        delay_table_creation: bool = True,
        enable_index: bool = False,
        th_create_ivfflat_index: int = _OCEANBASE_DEFAULT_IVFFLAT_ROW_THRESHOLD,
        sql_logger: Optional[logging.Logger] = None,
        collection_stat: Optional[OceanBaseCollectionStat] = None,
        enable_normalize_vector: bool = False,
    ) -> None:
        """Create OceanBase Vector Store instance."""
        self.database = database
        self.connection_string = connection_string
        self.embedding_function = embedding_function
        self.embedding_dimension = embedding_dimension
        self.collection_name = collection_name
        self.pre_delete_collection = pre_delete_collection
        self.logger = logger or logging.getLogger(__name__)
        self.delay_table_creation = delay_table_creation
        self.th_create_ivfflat_index = th_create_ivfflat_index
        self.enable_index = enable_index
        self.sql_logger = sql_logger
        self.collection_stat = collection_stat
        self.enable_normalize_vector = enable_normalize_vector
        self.__post_init__(engine_args)

    def __post_init__(
        self,
        engine_args: Optional[dict] = None,
    ) -> None:
        """Create connection & vector table."""
        _engine_args = engine_args or {}
        if "pool_recycle" not in _engine_args:
            _engine_args["pool_recycle"] = 3600
        self.engine = create_engine(self.connection_string, **_engine_args)
        self.create_collection()

    @property
    def embeddings(self) -> Embeddings:
        """Get embedding function."""
        return self.embedding_function

    def create_collection(self) -> None:
        """Create vector table."""
        if self.pre_delete_collection:
            self.delete_collection()
        if not self.delay_table_creation and (
            self.collection_stat is None
            or self.collection_stat.get_maybe_collection_not_exist()
        ):
            self.create_table_if_not_exists()
            if self.collection_stat is not None:
                self.collection_stat.collection_exists()

    def delete_collection(self) -> None:
        """Drop vector table."""
        drop_statement = text(f"DROP TABLE IF EXISTS {self.collection_name}")
        if self.sql_logger is not None:
            self.sql_logger.debug(f"Trying to delete collection: {drop_statement}")
        with self.engine.connect() as conn, conn.begin():
            conn.execute(drop_statement)
            if self.collection_stat is not None:
                self.collection_stat.collection_not_exists()
                self.collection_stat.collection_index_not_exists()

    def create_table_if_not_exists(self) -> None:
        """Create vector table with SQL."""
        create_table_query = f"""
            CREATE TABLE IF NOT EXISTS `{self.collection_name}` (
                id VARCHAR(40) NOT NULL,
                embedding VECTOR({self.embedding_dimension}),
                document LONGTEXT,
                metadata JSON,
                PRIMARY KEY (id)
            )
        """
        if self.sql_logger is not None:
            self.sql_logger.debug(f"Trying to create table: {create_table_query}")
        with self.engine.connect() as conn, conn.begin():
            # Create the table
            conn.execute(text(create_table_query))

    def create_collection_ivfflat_index_if_not_exists(self) -> None:
        """Create ivfflat index table with SQL."""
        create_index_query = f"""
            CREATE INDEX IF NOT EXISTS `embedding_idx` on `{self.collection_name}` (
                embedding l2
            ) using ivfflat with (lists=20)
        """
        with ob_grwlock.writer_lock(), self.engine.connect() as conn, conn.begin():
            # Create Ivfflat Index
            if self.sql_logger is not None:
                self.sql_logger.debug(
                    f"Trying to create ivfflat index: {create_index_query}"
                )
            conn.execute(text(create_index_query))

    def check_table_exists(self) -> bool:
        """Whether table `collection_name` exists."""
        check_table_query = f"""
            SELECT COUNT(*) as cnt
            FROM information_schema.tables
            WHERE table_schema='{self.database}' AND table_name='{self.collection_name}'
        """
        try:
            with self.engine.connect() as conn, conn.begin(), ob_grwlock.reader_lock():
                table_exists_res = conn.execute(text(check_table_query))
                for row in table_exists_res:
                    return row.cnt > 0
                # No `cnt` rows? Just return False to pass `make mypy`
                return False
        except Exception as e:
            logger.error(f"check_table_exists error: {e}")
            return False

    def add_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[List[dict]] = None,
        ids: Optional[List[str]] = None,
        batch_size: int = 500,
        **kwargs: Any,
    ) -> List[str]:
        """Insert texts into vector table."""
        if ids is None:
            ids = [str(uuid.uuid1()) for _ in texts]

        embeddings = self.embedding_function.embed_documents(list(texts))

        if len(embeddings) == 0:
            return ids

        if not metadatas:
            metadatas = [{} for _ in texts]

        if self.delay_table_creation and (
            self.collection_stat is None
            or self.collection_stat.get_maybe_collection_not_exist()
        ):
            self.embedding_dimension = len(embeddings[0])
            self.create_table_if_not_exists()
            self.delay_table_creation = False
            if self.collection_stat is not None:
                self.collection_stat.collection_exists()

        chunks_table = Table(
            self.collection_name,
            Base.metadata,
            Column("id", VARCHAR(40), primary_key=True),
            Column("embedding", Vector(self.embedding_dimension)),
            Column("document", LONGTEXT, nullable=True),
            Column("metadata", JSON, nullable=True),  # filter
            keep_existing=True,
        )

        row_count_query = f"""
            SELECT COUNT(*) as cnt FROM `{self.collection_name}`
        """
        chunks_table_data = []
        with self.engine.connect() as conn, conn.begin():
            for document, metadata, chunk_id, embedding in zip(
                texts, metadatas, ids, embeddings
            ):
                chunks_table_data.append(
                    {
                        "id": chunk_id,
                        "embedding": embedding
                        if not self.enable_normalize_vector
                        else self._normalization_vectors(embedding),
                        "document": document,
                        "metadata": metadata,
                    }
                )

                # Execute the batch insert when the batch size is reached
                if len(chunks_table_data) == batch_size:
                    with ob_grwlock.reader_lock():
                        if self.sql_logger is not None:
                            insert_sql_for_log = str(
                                insert(chunks_table).values(chunks_table_data)
                            )
                            self.sql_logger.debug(
                                f"""Trying to insert vectors:
                                    {insert_sql_for_log}"""
                            )
                        conn.execute(insert(chunks_table).values(chunks_table_data))
                    # Clear the chunks_table_data list for the next batch
                    chunks_table_data.clear()

            # Insert any remaining records that didn't make up a full batch
            if chunks_table_data:
                with ob_grwlock.reader_lock():
                    if self.sql_logger is not None:
                        insert_sql_for_log = str(
                            insert(chunks_table).values(chunks_table_data)
                        )
                        self.sql_logger.debug(
                            f"""Trying to insert vectors:
                                {insert_sql_for_log}"""
                        )
                    conn.execute(insert(chunks_table).values(chunks_table_data))

            if self.enable_index and (
                self.collection_stat is None
                or self.collection_stat.get_maybe_collection_index_not_exist()
            ):
                with ob_grwlock.reader_lock():
                    row_cnt_res = conn.execute(text(row_count_query))
                for row in row_cnt_res:
                    if row.cnt > self.th_create_ivfflat_index:
                        self.create_collection_ivfflat_index_if_not_exists()
                        if self.collection_stat is not None:
                            self.collection_stat.collection_index_exists()

        return ids

    def similarity_search(
        self,
        query: str,
        k: int = 4,
        filter: Optional[MetadataFilters] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Do similarity search via query in String."""
        embedding = self.embedding_function.embed_query(query)
        docs = self.similarity_search_by_vector(embedding=embedding, k=k, filter=filter)
        return docs

    def similarity_search_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        filter: Optional[MetadataFilters] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Do similarity search via query vector."""
        docs_and_scores = self.similarity_search_with_score_by_vector(
            embedding=embedding, k=k, filter=filter
        )
        return [doc for doc, _ in docs_and_scores]

    def similarity_search_with_score_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        filter: Optional[MetadataFilters] = None,
        score_threshold: Optional[float] = None,
    ) -> List[Tuple[Document, float]]:
        """Do similarity search via query vector with score."""
        try:
            from sqlalchemy.engine import Row
        except ImportError:
            raise ImportError(
                "Could not import Row from sqlalchemy.engine. "
                "Please 'pip install sqlalchemy>=1.4'."
            )

        # filter is not support in OceanBase currently.

        # normailze embedding vector
        if self.enable_normalize_vector:
            embedding = self._normalization_vectors(embedding)

        embedding_str = ob_vector_to_db(embedding, self.embedding_dimension)
        vector_distance_op = "<@>" if self.enable_normalize_vector else "<->"
        sql_query = f"""
            SELECT document, metadata, embedding {vector_distance_op} '{embedding_str}'
            as distance
            FROM {self.collection_name}
            ORDER BY embedding {vector_distance_op} '{embedding_str}'
            LIMIT :k
        """
        sql_query_str_for_log = f"""
            SELECT document, metadata, embedding {vector_distance_op} '?' as distance
            FROM {self.collection_name}
            ORDER BY embedding {vector_distance_op} '?'
            LIMIT {k}
        """

        params = {"k": k}
        try:
            with ob_grwlock.reader_lock(), self.engine.connect() as conn:
                if self.sql_logger is not None:
                    self.sql_logger.debug(
                        f"Trying to do similarity search: {sql_query_str_for_log}"
                    )
                results: Sequence[Row] = conn.execute(
                    text(sql_query), params
                ).fetchall()

            if (score_threshold is not None) and self.enable_normalize_vector:
                documents_with_scores = [
                    (
                        Document(
                            content=result.document,
                            metadata=json.loads(result.metadata),
                        ),
                        result.distance,
                    )
                    for result in results
                    if result.distance >= score_threshold
                ]
            else:
                documents_with_scores = [
                    (
                        Document(
                            content=result.document,
                            metadata=json.loads(result.metadata),
                        ),
                        result.distance,
                    )
                    for result in results
                ]
            return documents_with_scores
        except Exception as e:
            self.logger.error("similarity_search_with_score_by_vector failed:", str(e))
            return []

    def similarity_search_with_score(
        self,
        query: str,
        k: int = 4,
        filter: Optional[MetadataFilters] = None,
        score_threshold: Optional[float] = None,
    ) -> List[Tuple[Document, float]]:
        """Do similarity search via query String with score."""
        embedding = self.embedding_function.embed_query(query)
        docs = self.similarity_search_with_score_by_vector(
            embedding=embedding, k=k, filter=filter, score_threshold=score_threshold
        )
        return docs

    def delete(self, ids: Optional[List[str]] = None, **kwargs: Any) -> Optional[bool]:
        """Delete vectors from vector table."""
        if ids is None:
            raise ValueError("No ids provided to delete.")

        # Define the table schema
        chunks_table = Table(
            self.collection_name,
            Base.metadata,
            Column("id", VARCHAR(40), primary_key=True),
            Column("embedding", Vector(self.embedding_dimension)),
            Column("document", LONGTEXT, nullable=True),
            Column("metadata", JSON, nullable=True),  # filter
            keep_existing=True,
        )

        try:
            with self.engine.connect() as conn, conn.begin():
                delete_condition = chunks_table.c.id.in_(ids)
                delete_stmt = chunks_table.delete().where(delete_condition)
                with ob_grwlock.reader_lock():
                    if self.sql_logger is not None:
                        self.sql_logger.debug(
                            f"Trying to delete vectors: {str(delete_stmt)}"
                        )
                    conn.execute(delete_stmt)
                return True
        except Exception as e:
            self.logger.error("Delete operation failed:", str(e))
            return False

    def _normalization_vectors(self, vector):
        import numpy as np

        norm = np.linalg.norm(vector)
        return (vector / norm).tolist()

    @classmethod
    def connection_string_from_db_params(
        cls,
        host: str,
        port: int,
        database: str,
        user: str,
        password: str,
    ) -> str:
        """Get connection string."""
        return f"mysql+pymysql://{user}:{password}@{host}:{port}/{database}"


ob_collection_stats_lock = threading.Lock()
ob_collection_stats: Dict[str, OceanBaseCollectionStat] = {}


@register_resource(
    _("OceanBase Vector Store"),
    "oceanbase_vector_store",
    category=ResourceCategory.VECTOR_STORE,
    parameters=[
        *_COMMON_PARAMETERS,
        Parameter.build_from(
            _("OceanBase Host"),
            "ob_host",
            str,
            description=_("oceanbase host"),
            optional=True,
            default=None,
        ),
        Parameter.build_from(
            _("OceanBase Port"),
            "ob_port",
            int,
            description=_("oceanbase port"),
            optional=True,
            default=None,
        ),
        Parameter.build_from(
            _("OceanBase User"),
            "ob_user",
            str,
            description=_("user to login"),
            optional=True,
            default=None,
        ),
        Parameter.build_from(
            _("OceanBase Password"),
            "ob_password",
            str,
            description=_("password to login"),
            optional=True,
            default=None,
        ),
        Parameter.build_from(
            _("OceanBase Database"),
            "ob_database",
            str,
            description=_("database for vector tables"),
            optional=True,
            default=None,
        ),
    ],
    description="OceanBase vector store.",
)
class OceanBaseConfig(VectorStoreConfig):
    """OceanBase vector store config."""

    class Config:
        """Config for BaseModel."""

        arbitrary_types_allowed = True

    """OceanBase config"""
    ob_host: str = Field(
        default="127.0.0.1",
        description="oceanbase host",
    )
    ob_port: int = Field(
        default=2881,
        description="oceanbase port",
    )
    ob_user: str = Field(
        default="root@test",
        description="user to login",
    )
    ob_password: str = Field(
        default="",
        description="password to login",
    )
    ob_database: str = Field(
        default="test",
        description="database for vector tables",
    )


class OceanBaseStore(VectorStoreBase):
    """OceanBase vector store."""

    def __init__(self, vector_store_config: OceanBaseConfig) -> None:
        """Create a OceanBaseStore instance."""
        if vector_store_config.embedding_fn is None:
            raise ValueError("embedding_fn is required for OceanBaseStore")

        self.embeddings = vector_store_config.embedding_fn
        self.collection_name = vector_store_config.name
        vector_store_config = vector_store_config.dict()
        self.OB_HOST = str(
            vector_store_config.get("ob_host") or os.getenv("OB_HOST", "127.0.0.1")
        )
        self.OB_PORT = int(
            vector_store_config.get("ob_port") or int(os.getenv("OB_PORT", "2881"))
        )
        self.OB_USER = str(
            vector_store_config.get("ob_user") or os.getenv("OB_USER", "root@test")
        )
        self.OB_PASSWORD = str(
            vector_store_config.get("ob_password") or os.getenv("OB_PASSWORD", "")
        )
        self.OB_DATABASE = str(
            vector_store_config.get("ob_database") or os.getenv("OB_DATABASE", "test")
        )
        self.OB_ENABLE_NORMALIZE_VECTOR = bool(
            os.getenv("OB_ENABLE_NORMALIZE_VECTOR", "")
        )
        self.connection_string = OceanBase.connection_string_from_db_params(
            self.OB_HOST, self.OB_PORT, self.OB_DATABASE, self.OB_USER, self.OB_PASSWORD
        )
        self.logger = logger
        with ob_collection_stats_lock:
            if ob_collection_stats.get(self.collection_name) is None:
                ob_collection_stats[self.collection_name] = OceanBaseCollectionStat()
            self.collection_stat = ob_collection_stats[self.collection_name]

        self.vector_store_client = OceanBase(
            database=self.OB_DATABASE,
            connection_string=self.connection_string,
            embedding_function=self.embeddings,
            collection_name=self.collection_name,
            logger=self.logger,
            sql_logger=sql_logger,
            enable_index=bool(os.getenv("OB_ENABLE_INDEX", "")),
            collection_stat=self.collection_stat,
            enable_normalize_vector=self.OB_ENABLE_NORMALIZE_VECTOR,
        )

    def similar_search(
        self, text, topk, filters: Optional[MetadataFilters] = None, **kwargs: Any
    ) -> List[Chunk]:
        """Perform a search on a query string and return results."""
        self.logger.info("OceanBase: similar_search..")
        documents = self.vector_store_client.similarity_search(
            text, topk, filter=filters
        )
        return [Chunk(content=doc.content, metadata=doc.metadata) for doc in documents]

    def similar_search_with_scores(
        self,
        text,
        topk,
        score_threshold: float,
        filters: Optional[MetadataFilters] = None,
    ) -> List[Chunk]:
        """Perform a search on a query string and return results with score."""
        self.logger.info("OceanBase: similar_search_with_scores..")
        docs_and_scores = self.vector_store_client.similarity_search_with_score(
            text, topk, filter=filters
        )
        return [
            Chunk(content=doc.content, metadata=doc.metadata, score=score)
            for doc, score in docs_and_scores
        ]

    def vector_name_exists(self):
        """Whether vector name exists."""
        self.logger.info("OceanBase: vector_name_exists..")
        return self.vector_store_client.check_table_exists()

    def load_document(self, chunks: List[Chunk]) -> List[str]:
        """Load document in vector database."""
        self.logger.info("OceanBase: load_document..")
        # lc_documents = [Chunk.chunk2langchain(chunk) for chunk in chunks]
        texts = [chunk.content for chunk in chunks]
        metadatas = [chunk.metadata for chunk in chunks]
        ids = self.vector_store_client.add_texts(texts=texts, metadatas=metadatas)
        return ids

    def delete_vector_name(self, vector_name):
        """Delete vector name."""
        self.logger.info("OceanBase: delete_vector_name..")
        return self.vector_store_client.delete_collection()

    def delete_by_ids(self, ids):
        """Delete vector by ids."""
        self.logger.info("OceanBase: delete_by_ids..")
        ids = ids.split(",")
        if len(ids) > 0:
            self.vector_store_client.delete(ids)



"""Elasticsearch vector store."""

from __future__ import annotations

import logging
import os
from typing import List, Optional

from dbgpt._private.pydantic import Field
from dbgpt.core import Chunk, Embeddings
from dbgpt.core.awel.flow import Parameter, ResourceCategory, register_resource
from dbgpt.storage.vector_store.base import (
    _COMMON_PARAMETERS,
    VectorStoreBase,
    VectorStoreConfig,
)
from dbgpt.storage.vector_store.filters import MetadataFilters
from dbgpt.util import string_utils
from dbgpt.util.i18n_utils import _

logger = logging.getLogger(__name__)


@register_resource(
    _("ElasticSearch Vector Store"),
    "elasticsearch_vector_store",
    category=ResourceCategory.VECTOR_STORE,
    parameters=[
        *_COMMON_PARAMETERS,
        Parameter.build_from(
            _("Uri"),
            "uri",
            str,
            description=_(
                "The uri of elasticsearch store, if not set, will use the default "
                "uri."
            ),
            optional=True,
            default="localhost",
        ),
        Parameter.build_from(
            _("Port"),
            "port",
            str,
            description=_(
                "The port of elasticsearch store, if not set, will use the default "
                "port."
            ),
            optional=True,
            default="9200",
        ),
        Parameter.build_from(
            _("Alias"),
            "alias",
            str,
            description=_(
                "The alias of elasticsearch store, if not set, will use the default "
                "alias."
            ),
            optional=True,
            default="default",
        ),
        Parameter.build_from(
            _("Index Name"),
            "index_name",
            str,
            description=_(
                "The index name of elasticsearch store, if not set, will use the "
                "default index name."
            ),
            optional=True,
            default="index_name_test",
        ),
    ],
    description=_("Elasticsearch vector store."),
)
class ElasticsearchVectorConfig(VectorStoreConfig):
    """Elasticsearch vector store config."""

    class Config:
        """Config for BaseModel."""

        arbitrary_types_allowed = True

    uri: str = Field(
        default="localhost",
        description="The uri of elasticsearch store, if not set, will use the default "
        "uri.",
    )
    port: str = Field(
        default="9200",
        description="The port of elasticsearch store, if not set, will use the default "
        "port.",
    )

    alias: str = Field(
        default="default",
        description="The alias of elasticsearch store, if not set, will use the "
        "default "
        "alias.",
    )
    index_name: str = Field(
        default="index_name_test",
        description="The index name of elasticsearch store, if not set, will use the "
        "default index name.",
    )
    metadata_field: str = Field(
        default="metadata",
        description="The metadata field of elasticsearch store, if not set, will use "
        "the default metadata field.",
    )
    secure: str = Field(
        default="",
        description="The secure of elasticsearch store, if not set, will use the "
        "default secure.",
    )


class ElasticStore(VectorStoreBase):
    """Elasticsearch vector store."""

    def __init__(self, vector_store_config: ElasticsearchVectorConfig) -> None:
        """Create a ElasticsearchStore instance.

        Args:
            vector_store_config (ElasticsearchVectorConfig): ElasticsearchStore config.
        """
        connect_kwargs = {}
        elasticsearch_vector_config = vector_store_config.dict()
        self.uri = elasticsearch_vector_config.get("uri") or os.getenv(
            "ELASTICSEARCH_URL", "localhost"
        )
        self.port = elasticsearch_vector_config.get("post") or os.getenv(
            "ELASTICSEARCH_PORT", "9200"
        )
        self.username = elasticsearch_vector_config.get("username") or os.getenv(
            "ELASTICSEARCH_USERNAME"
        )
        self.password = elasticsearch_vector_config.get("password") or os.getenv(
            "ELASTICSEARCH_PASSWORD"
        )

        self.collection_name = (
            elasticsearch_vector_config.get("name") or vector_store_config.name
        )
        # name to hex
        if string_utils.is_all_chinese(self.collection_name):
            bytes_str = self.collection_name.encode("utf-8")
            hex_str = bytes_str.hex()
            self.collection_name = hex_str
        if vector_store_config.embedding_fn is None:
            # Perform runtime checks on self.embedding to
            # ensure it has been correctly set and loaded
            raise ValueError("embedding_fn is required for ElasticSearchStore")
        # to lower case
        self.index_name = self.collection_name.lower()
        self.embedding: Embeddings = vector_store_config.embedding_fn
        self.fields: List = []

        if (self.username is None) != (self.password is None):
            raise ValueError(
                "Both username and password must be set to use authentication for "
                "ElasticSearch"
            )

        if self.username:
            connect_kwargs["username"] = self.username
            connect_kwargs["password"] = self.password

        # english index settings
        self.index_settings = {
            "settings": {
                "number_of_shards": 1,
                "number_of_replicas": 0,  # replica number
            }
        }
        """"""
        try:
            from elasticsearch import Elasticsearch
            from langchain.vectorstores.elasticsearch import ElasticsearchStore
        except ImportError:
            raise ValueError(
                "Could not import langchain and elasticsearch python package. "
                "Please install it with `pip install langchain` and "
                "`pip install elasticsearch`."
            )
        try:
            if self.username != "" and self.password != "":
                self.es_client_python = Elasticsearch(
                    f"http://{self.uri}:{self.port}",
                    basic_auth=(self.username, self.password),
                )
                # create es index
                if not self.vector_name_exists():
                    self.es_client_python.indices.create(
                        index=self.index_name, body=self.index_settings
                    )
            else:
                logger.warning("ElasticSearch not set username and password")
                self.es_client_python = Elasticsearch(f"http://{self.uri}:{self.port}")
                if not self.vector_name_exists():
                    self.es_client_python.indices.create(
                        index=self.index_name, body=self.index_settings
                    )
        except ConnectionError:
            logger.error("ElasticSearch connection failed")
        except Exception as e:
            logger.error(f"ElasticSearch connection failed : {e}")

        # create es index
        try:
            if self.username != "" and self.password != "":
                self.db_init = ElasticsearchStore(
                    es_url=f"http://{self.uri}:{self.port}",
                    index_name=self.index_name,
                    query_field="context",
                    vector_query_field="dense_vector",
                    embedding=self.embedding,  # type: ignore
                    es_user=self.username,
                    es_password=self.password,
                )
            else:
                logger.warning("ElasticSearch not set username and password")
                self.db_init = ElasticsearchStore(
                    es_url=f"http://{self.uri}:{self.port}",
                    index_name=self.index_name,
                    query_field="context",
                    vector_query_field="dense_vector",
                    embedding=self.embedding,  # type: ignore
                )
        except ConnectionError:
            logger.error("ElasticSearch connection failed")
        except Exception as e:
            logger.error(f"ElasticSearch connection failed: {e}")

    def load_document(
        self,
        chunks: List[Chunk],
    ) -> List[str]:
        """Add text data into ElasticSearch."""
        logger.info("ElasticStore load document")
        try:
            from langchain.vectorstores.elasticsearch import ElasticsearchStore
        except ImportError:
            raise ValueError(
                "Could not import langchain python package. "
                "Please install it with `pip install langchain` and "
                "`pip install elasticsearch`."
            )
        try:
            texts = [chunk.content for chunk in chunks]
            metadatas = [chunk.metadata for chunk in chunks]
            ids = [chunk.chunk_id for chunk in chunks]
            if self.username != "" and self.password != "":
                self.db = ElasticsearchStore.from_texts(
                    texts=texts,
                    embedding=self.embedding,  # type: ignore
                    metadatas=metadatas,
                    ids=ids,
                    es_url=f"http://{self.uri}:{self.port}",
                    index_name=self.index_name,
                    # Defaults to COSINE. Can be one of COSINE, EUCLIDEAN_DISTANCE
                    # , or DOT_PRODUCT.
                    distance_strategy="COSINE",
                    # Name of the field to store the texts in.
                    query_field="context",
                    # Optional. Name of the field to store the embedding vectors in.
                    vector_query_field="dense_vector",
                    # verify_certs=False,
                    # strategy: Optional. Retrieval strategy to use when searching the
                    # index.
                    # Defaults to ApproxRetrievalStrategy.
                    # Can be one of ExactRetrievalStrategy, ApproxRetrievalStrategy,
                    # or SparseRetrievalStrategy.
                    es_user=self.username,
                    es_password=self.password,
                )  # type: ignore
                logger.info("Elasticsearch save success.......")
                return ids
            else:
                self.db = ElasticsearchStore.from_documents(
                    texts=texts,
                    embedding=self.embedding,  # type: ignore
                    metadatas=metadatas,
                    ids=ids,
                    es_url=f"http://{self.uri}:{self.port}",
                    index_name=self.index_name,
                    distance_strategy="COSINE",
                    query_field="context",
                    vector_query_field="dense_vector",
                    # verify_certs=False,
                )  # type: ignore
                return ids
        except ConnectionError as ce:
            logger.error(f"ElasticSearch connect failed {ce}")
        except Exception as e:
            logger.error(f"ElasticSearch load_document failed : {e}")
        return []

    def delete_by_ids(self, ids):
        """Delete vector by ids."""
        logger.info(f"begin delete elasticsearch len ids: {len(ids)}")
        ids = ids.split(",")
        try:
            self.db_init.delete(ids=ids)
            self.es_client_python.indices.refresh(index=self.index_name)
        except Exception as e:
            logger.error(f"ElasticSearch delete_by_ids failed : {e}")

    def similar_search(
        self,
        text: str,
        topk: int,
        filters: Optional[MetadataFilters] = None,
    ) -> List[Chunk]:
        """Perform a search on a query string and return results."""
        info_docs = self._search(query=text, topk=topk, filters=filters)
        return info_docs

    def similar_search_with_scores(
        self, text, topk, score_threshold, filters: Optional[MetadataFilters] = None
    ) -> List[Chunk]:
        """Perform a search on a query string and return results with score.

        For more information about the search parameters, take a look at the
        ElasticSearch documentation found here: https://www.elastic.co/.

        Args:
            text (str): The query text.
            topk (int): The number of similar documents to return.
            score_threshold (float): Optional, a floating point value between 0 to 1.
            filters (Optional[MetadataFilters]): Optional, metadata filters.
        Returns:
            List[Chunk]: Result doc and score.
        """
        query = text
        info_docs = self._search(query=query, topk=topk, filters=filters)
        docs_and_scores = [
            chunk for chunk in info_docs if chunk.score >= score_threshold
        ]
        if len(docs_and_scores) == 0:
            logger.warning(
                "No relevant docs were retrieved using the relevance score"
                f" threshold {score_threshold}"
            )
        return docs_and_scores

    def _search(
        self, query: str, topk: int, filters: Optional[MetadataFilters] = None, **kwargs
    ) -> List[Chunk]:
        """Search similar documents.

        Args:
            query: query text
            topk: return docs nums. Defaults to 4.
            filters: metadata filters.
        Return:
            List[Chunk]: list of chunks
        """
        jieba_tokenize = kwargs.pop("jieba_tokenize", None)
        if jieba_tokenize:
            try:
                import jieba
                import jieba.analyse
            except ImportError:
                raise ValueError("Please install it with `pip install jieba`.")
            query_list = jieba.analyse.textrank(query, topK=20, withWeight=False)
            query = " ".join(query_list)
        body = {"query": {"match": {"context": query}}}
        search_results = self.es_client_python.search(
            index=self.index_name, body=body, size=topk
        )
        search_results = search_results["hits"]["hits"]

        if not search_results:
            logger.warning("""No ElasticSearch results found.""")
            return []
        info_docs = []
        for result in search_results:
            doc_id = result["_id"]
            source = result["_source"]
            context = source["context"]
            metadata = source["metadata"]
            score = result["_score"]
            doc_with_score = Chunk(
                content=context, metadata=metadata, score=score, chunk_id=doc_id
            )
            info_docs.append(doc_with_score)
        return info_docs

    def vector_name_exists(self):
        """Whether vector name exists."""
        return self.es_client_python.indices.exists(index=self.index_name)

    def delete_vector_name(self, vector_name: str):
        """Delete vector name/index_name."""
        if self.es_client_python.indices.exists(index=self.index_name):
            self.es_client_python.indices.delete(index=self.index_name)



"""Vector store base class."""
import logging
import math
from abc import ABC, abstractmethod
from typing import Any, List, Optional

from dbgpt._private.pydantic import ConfigDict, Field
from dbgpt.core import Chunk, Embeddings
from dbgpt.core.awel.flow import Parameter
from dbgpt.rag.index.base import IndexStoreBase, IndexStoreConfig
from dbgpt.storage.vector_store.filters import MetadataFilters
from dbgpt.util.i18n_utils import _

logger = logging.getLogger(__name__)

_COMMON_PARAMETERS = [
    Parameter.build_from(
        _("Collection Name"),
        "name",
        str,
        description=_(
            "The name of vector store, if not set, will use the default name."
        ),
        optional=True,
        default="dbgpt_collection",
    ),
    Parameter.build_from(
        _("User"),
        "user",
        str,
        description=_(
            "The user of vector store, if not set, will use the default user."
        ),
        optional=True,
        default=None,
    ),
    Parameter.build_from(
        _("Password"),
        "password",
        str,
        description=_(
            "The password of vector store, if not set, will use the "
            "default password."
        ),
        optional=True,
        default=None,
    ),
    Parameter.build_from(
        _("Embedding Function"),
        "embedding_fn",
        Embeddings,
        description=_(
            "The embedding function of vector store, if not set, will use "
            "the default embedding function."
        ),
        optional=True,
        default=None,
    ),
    Parameter.build_from(
        _("Max Chunks Once Load"),
        "max_chunks_once_load",
        int,
        description=_(
            "The max number of chunks to load at once. If your document is "
            "large, you can set this value to a larger number to speed up the loading "
            "process. Default is 10."
        ),
        optional=True,
        default=10,
    ),
    Parameter.build_from(
        _("Max Threads"),
        "max_threads",
        int,
        description=_(
            "The max number of threads to use. Default is 1. If you set "
            "this bigger than 1, please make sure your vector store is thread-safe."
        ),
        optional=True,
        default=1,
    ),
]


class VectorStoreConfig(IndexStoreConfig):
    """Vector store config."""

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    user: Optional[str] = Field(
        default=None,
        description="The user of vector store, if not set, will use the default user.",
    )
    password: Optional[str] = Field(
        default=None,
        description=(
            "The password of vector store, if not set, will use the default password."
        ),
    )


class VectorStoreBase(IndexStoreBase, ABC):
    """Vector store base class."""

    def filter_by_score_threshold(
        self, chunks: List[Chunk], score_threshold: float
    ) -> List[Chunk]:
        """Filter chunks by score threshold.

        Args:
            chunks(List[Chunks]): The chunks to filter.
            score_threshold(float): The score threshold.
        Return:
            List[Chunks]: The filtered chunks.
        """
        candidates_chunks = chunks
        if score_threshold is not None:
            candidates_chunks = [
                Chunk(
                    metadata=chunk.metadata,
                    content=chunk.content,
                    score=chunk.score,
                    chunk_id=str(id),
                )
                for chunk in chunks
                if chunk.score >= score_threshold
            ]
            if len(candidates_chunks) == 0:
                logger.warning(
                    "No relevant docs were retrieved using the relevance score"
                    f" threshold {score_threshold}"
                )
        return candidates_chunks

    @abstractmethod
    def vector_name_exists(self) -> bool:
        """Whether vector name exists."""
        return False

    def convert_metadata_filters(self, filters: MetadataFilters) -> Any:
        """Convert metadata filters to vector store filters.

        Args:
            filters: (Optional[MetadataFilters]) metadata filters.
        """
        raise NotImplementedError

    def _normalization_vectors(self, vectors):
        """Return L2-normalization vectors to scale[0,1].

        Normalization vectors to scale[0,1].
        """
        import numpy as np

        norm = np.linalg.norm(vectors)
        return vectors / norm

    def _default_relevance_score_fn(self, distance: float) -> float:
        """Return a similarity score on a scale [0, 1]."""
        return 1.0 - distance / math.sqrt(2)

    async def aload_document(self, chunks: List[Chunk]) -> List[str]:  # type: ignore
        """Load document in index database.

        Args:
            chunks(List[Chunk]): document chunks.

        Return:
            List[str]: chunk ids.
        """
        raise NotImplementedError



"""Vector Store Module."""
from typing import Tuple, Type


def _import_pgvector() -> Tuple[Type, Type]:
    from dbgpt.storage.vector_store.pgvector_store import PGVectorConfig, PGVectorStore

    return PGVectorStore, PGVectorConfig


def _import_milvus() -> Tuple[Type, Type]:
    from dbgpt.storage.vector_store.milvus_store import MilvusStore, MilvusVectorConfig

    return MilvusStore, MilvusVectorConfig


def _import_chroma() -> Tuple[Type, Type]:
    from dbgpt.storage.vector_store.chroma_store import ChromaStore, ChromaVectorConfig

    return ChromaStore, ChromaVectorConfig


def _import_weaviate() -> Tuple[Type, Type]:
    from dbgpt.storage.vector_store.weaviate_store import (
        WeaviateStore,
        WeaviateVectorConfig,
    )

    return WeaviateStore, WeaviateVectorConfig


def _import_oceanbase() -> Tuple[Type, Type]:
    from dbgpt.storage.vector_store.oceanbase_store import (
        OceanBaseConfig,
        OceanBaseStore,
    )

    return OceanBaseStore, OceanBaseConfig


def _import_elastic() -> Tuple[Type, Type]:
    from dbgpt.storage.vector_store.elastic_store import (
        ElasticsearchVectorConfig,
        ElasticStore,
    )

    return ElasticStore, ElasticsearchVectorConfig


def _import_builtin_knowledge_graph() -> Tuple[Type, Type]:
    from dbgpt.storage.knowledge_graph.knowledge_graph import (
        BuiltinKnowledgeGraph,
        BuiltinKnowledgeGraphConfig,
    )

    return BuiltinKnowledgeGraph, BuiltinKnowledgeGraphConfig


def _import_openspg() -> Tuple[Type, Type]:
    from dbgpt.storage.knowledge_graph.open_spg import OpenSPG, OpenSPGConfig

    return OpenSPG, OpenSPGConfig


def __getattr__(name: str) -> Tuple[Type, Type]:
    if name == "Chroma":
        return _import_chroma()
    elif name == "Milvus":
        return _import_milvus()
    elif name == "Weaviate":
        return _import_weaviate()
    elif name == "PGVector":
        return _import_pgvector()
    elif name == "OceanBase":
        return _import_oceanbase()
    elif name == "ElasticSearch":
        return _import_elastic()
    elif name == "KnowledgeGraph":
        return _import_builtin_knowledge_graph()
    elif name == "OpenSPG":
        return _import_openspg()
    else:
        raise AttributeError(f"Could not find: {name}")


__vector_store__ = [
    "Chroma",
    "Milvus",
    "Weaviate",
    "OceanBase",
    "PGVector",
    "ElasticSearch",
]

__knowledge_graph__ = ["KnowledgeGraph", "OpenSPG"]

__all__ = __vector_store__ + __knowledge_graph__



"""LLM module."""
import logging
from collections import defaultdict
from enum import Enum
from typing import Any, Dict, List, Optional, Type

from dbgpt._private.pydantic import BaseModel, ConfigDict, Field
from dbgpt.core import LLMClient, ModelMetadata, ModelRequest

logger = logging.getLogger(__name__)


def _build_model_request(input_value: Dict) -> ModelRequest:
    """Build model request from input value.

    Args:
        input_value(str or dict): input value

    Returns:
        ModelRequest: model request, pass to llm client
    """
    parm = {
        "model": input_value.get("model"),
        "messages": input_value.get("messages"),
        "temperature": input_value.get("temperature", None),
        "max_new_tokens": input_value.get("max_new_tokens", None),
        "stop": input_value.get("stop", None),
        "stop_token_ids": input_value.get("stop_token_ids", None),
        "context_len": input_value.get("context_len", None),
        "echo": input_value.get("echo", None),
        "span_id": input_value.get("span_id", None),
    }

    return ModelRequest(**parm)


class LLMStrategyType(Enum):
    """LLM strategy type."""

    Priority = "priority"
    Auto = "auto"
    Default = "default"


class LLMStrategy:
    """LLM strategy base class."""

    def __init__(self, llm_client: LLMClient, context: Optional[str] = None):
        """Create an LLMStrategy instance."""
        self._llm_client = llm_client
        self._context = context

    @property
    def type(self) -> LLMStrategyType:
        """Return the strategy type."""
        return LLMStrategyType.Default

    def _excluded_models(
        self,
        all_models: List[ModelMetadata],
        excluded_models: List[str],
        need_uses: Optional[List[str]] = None,
    ):
        if not need_uses:
            need_uses = []
        can_uses = []
        for item in all_models:
            if item.model in need_uses and item.model not in excluded_models:
                can_uses.append(item)
        return can_uses

    async def next_llm(self, excluded_models: Optional[List[str]] = None):
        """Return next available llm model name.

        Args:
            excluded_models(List[str]): excluded models

        Returns:
            str: Next available llm model name
        """
        if not excluded_models:
            excluded_models = []
        try:
            all_models = await self._llm_client.models()
            available_llms = self._excluded_models(all_models, excluded_models, None)
            if available_llms and len(available_llms) > 0:
                return available_llms[0].model
            else:
                raise ValueError("No model service available!")

        except Exception as e:
            logger.error(f"{self.type} get next llm failed!{str(e)}")
            raise ValueError(f"Failed to allocate model service,{str(e)}!")


llm_strategies: Dict[LLMStrategyType, List[Type[LLMStrategy]]] = defaultdict(list)


def register_llm_strategy(
    llm_strategy_type: LLMStrategyType, strategy: Type[LLMStrategy]
):
    """Register llm strategy."""
    llm_strategies[llm_strategy_type].append(strategy)


class LLMConfig(BaseModel):
    """LLM configuration."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    llm_client: Optional[LLMClient] = Field(default_factory=LLMClient)
    llm_strategy: LLMStrategyType = Field(default=LLMStrategyType.Default)
    strategy_context: Optional[Any] = None



from .config_utils import AppConfig
from .pagination_utils import PaginationResult
from .parameter_utils import BaseParameters, EnvArgumentParser, ParameterDescription
from .utils import get_gpu_memory, get_or_create_event_loop

__ALL__ = [
    "get_gpu_memory",
    "get_or_create_event_loop",
    "PaginationResult",
    "BaseParameters",
    "ParameterDescription",
    "EnvArgumentParser",
    "AppConfig",
]



"""SQLAlchemy data types for StarRocks."""

import logging
import re
from typing import Any, Dict, List, Optional, Type

from sqlalchemy import Float, Integer, Numeric
from sqlalchemy.sql import sqltypes
from sqlalchemy.sql.type_api import TypeEngine

logger = logging.getLogger(__name__)


class TINYINT(Integer):  # pylint: disable=no-init
    """StarRocks TINYINT type."""

    __visit_name__ = "TINYINT"


class LARGEINT(Integer):  # pylint: disable=no-init
    """StarRocks LARGEINT type."""

    __visit_name__ = "LARGEINT"


class DOUBLE(Float):  # pylint: disable=no-init
    """StarRocks DOUBLE type."""

    __visit_name__ = "DOUBLE"


class HLL(Numeric):  # pylint: disable=no-init
    """StarRocks HLL type."""

    __visit_name__ = "HLL"


class BITMAP(Numeric):  # pylint: disable=no-init
    """StarRocks BITMAP type."""

    __visit_name__ = "BITMAP"


class PERCENTILE(Numeric):  # pylint: disable=no-init
    """StarRocks PERCENTILE type."""

    __visit_name__ = "PERCENTILE"


class ARRAY(TypeEngine):  # pylint: disable=no-init
    """StarRocks ARRAY type."""

    __visit_name__ = "ARRAY"

    @property
    def python_type(self) -> Optional[Type[List[Any]]]:  # type: ignore
        """Return the Python type for this SQL type."""
        return list


class MAP(TypeEngine):  # pylint: disable=no-init
    """StarRocks MAP type."""

    __visit_name__ = "MAP"

    @property
    def python_type(self) -> Optional[Type[Dict[Any, Any]]]:  # type: ignore
        """Return the Python type for this SQL type."""
        return dict


class STRUCT(TypeEngine):  # pylint: disable=no-init
    """StarRocks STRUCT type."""

    __visit_name__ = "STRUCT"

    @property
    def python_type(self) -> Optional[Type[Any]]:  # type: ignore
        """Return the Python type for this SQL type."""
        return None


_type_map = {
    # === Boolean ===
    "boolean": sqltypes.BOOLEAN,
    # === Integer ===
    "tinyint": sqltypes.SMALLINT,
    "smallint": sqltypes.SMALLINT,
    "int": sqltypes.INTEGER,
    "bigint": sqltypes.BIGINT,
    "largeint": LARGEINT,
    # === Floating-point ===
    "float": sqltypes.FLOAT,
    "double": DOUBLE,
    # === Fixed-precision ===
    "decimal": sqltypes.DECIMAL,
    # === String ===
    "varchar": sqltypes.VARCHAR,
    "char": sqltypes.CHAR,
    "json": sqltypes.JSON,
    # === Date and time ===
    "date": sqltypes.DATE,
    "datetime": sqltypes.DATETIME,
    "timestamp": sqltypes.DATETIME,
    # === Structural ===
    "array": ARRAY,
    "map": MAP,
    "struct": STRUCT,
    "hll": HLL,
    "percentile": PERCENTILE,
    "bitmap": BITMAP,
}


def parse_sqltype(type_str: str) -> TypeEngine:
    """Parse a SQL type string into a SQLAlchemy type object."""
    type_str = type_str.strip().lower()
    match = re.match(r"^(?P<type>\w+)\s*(?:\((?P<options>.*)\))?", type_str)
    if not match:
        logger.warning(f"Could not parse type name '{type_str}'")
        return sqltypes.NULLTYPE
    type_name = match.group("type")

    if type_name not in _type_map:
        logger.warning(f"Did not recognize type '{type_name}'")
        return sqltypes.NULLTYPE
    type_class = _type_map[type_name]
    return type_class()



"""DB-GPT Multi-Agents Module."""

from .core.action import *  # noqa: F401, F403
from .core.agent import (  # noqa: F401
    Agent,
    AgentContext,
    AgentGenerateContext,
    AgentMessage,
)
from .core.agent_manage import (  # noqa: F401
    AgentManager,
    get_agent_manager,
    initialize_agent,
)
from .core.base_agent import ConversableAgent  # noqa: F401
from .core.memory import *  # noqa: F401, F403
from .core.memory.gpts.gpts_memory import GptsMemory  # noqa: F401
from .core.plan import *  # noqa: F401, F403
from .core.profile import *  # noqa: F401, F403
from .core.schema import PluginStorageType  # noqa: F401
from .core.user_proxy_agent import UserProxyAgent  # noqa: F401
from .resource.base import AgentResource, Resource, ResourceType  # noqa: F401
from .util.llm.llm import LLMConfig  # noqa: F401

__ALL__ = [
    "Agent",
    "AgentContext",
    "AgentGenerateContext",
    "AgentMessage",
    "AgentManager",
    "initialize_agent",
    "get_agent_manager",
    "ConversableAgent",
    "Action",
    "ActionOutput",
    "LLMConfig",
    "GptsMemory",
    "AgentResource",
    "ResourceType",
    "PluginStorageType",
    "UserProxyAgent",
]



"""Hybrid memory module.

This structure explicitly models the human short-term and long-term memories. The
short-term memory temporarily buffers recent perceptions, while long-term memory
consolidates important information over time.
"""

import os.path
from concurrent.futures import Executor, ThreadPoolExecutor
from datetime import datetime
from typing import TYPE_CHECKING, Generic, List, Optional, Tuple, Type

from dbgpt.core import Embeddings, LLMClient
from dbgpt.util.annotations import immutable, mutable

from .base import (
    DiscardedMemoryFragments,
    ImportanceScorer,
    InsightExtractor,
    Memory,
    SensoryMemory,
    ShortTermMemory,
    T,
    WriteOperation,
)
from .long_term import LongTermMemory
from .short_term import EnhancedShortTermMemory

if TYPE_CHECKING:
    from dbgpt.storage.vector_store.connector import VectorStoreConnector


class HybridMemory(Memory, Generic[T]):
    """Hybrid memory for the agent."""

    importance_weight: float = 0.9

    def __init__(
        self,
        now: datetime,
        sensory_memory: SensoryMemory[T],
        short_term_memory: ShortTermMemory[T],
        long_term_memory: LongTermMemory[T],
        default_insight_extractor: Optional[InsightExtractor] = None,
        default_importance_scorer: Optional[ImportanceScorer] = None,
    ):
        """Create a hybrid memory."""
        self.now = now
        self._sensory_memory = sensory_memory
        self._short_term_memory = short_term_memory
        self._long_term_memory = long_term_memory
        self._default_insight_extractor = default_insight_extractor
        self._default_importance_scorer = default_importance_scorer

    def structure_clone(
        self: "HybridMemory[T]", now: Optional[datetime] = None
    ) -> "HybridMemory[T]":
        """Return a structure clone of the memory."""
        now = now or self.now
        m = HybridMemory(
            now=now,
            sensory_memory=self._sensory_memory.structure_clone(now),
            short_term_memory=self._short_term_memory.structure_clone(now),
            long_term_memory=self._long_term_memory.structure_clone(now),
        )
        m._copy_from(self)
        return m

    @classmethod
    def from_chroma(
        cls,
        vstore_name: Optional[str] = "_chroma_agent_memory_",
        vstore_path: Optional[str] = None,
        embeddings: Optional[Embeddings] = None,
        executor: Optional[Executor] = None,
        now: Optional[datetime] = None,
        sensory_memory: Optional[SensoryMemory[T]] = None,
        short_term_memory: Optional[ShortTermMemory[T]] = None,
        long_term_memory: Optional[LongTermMemory[T]] = None,
        **kwargs
    ):
        """Create a hybrid memory from Chroma vector store."""
        from dbgpt.configs.model_config import DATA_DIR
        from dbgpt.storage.vector_store.chroma_store import ChromaVectorConfig
        from dbgpt.storage.vector_store.connector import VectorStoreConnector

        if not embeddings:
            from dbgpt.rag.embedding import DefaultEmbeddingFactory

            embeddings = DefaultEmbeddingFactory.openai()

        vstore_path = vstore_path or os.path.join(DATA_DIR, "agent_memory")

        vector_store_connector = VectorStoreConnector.from_default(
            vector_store_type="Chroma",
            embedding_fn=embeddings,
            vector_store_config=ChromaVectorConfig(
                name=vstore_name,
                persist_path=vstore_path,
            ),
        )
        return cls.from_vstore(
            vector_store_connector=vector_store_connector,
            embeddings=embeddings,
            executor=executor,
            now=now,
            sensory_memory=sensory_memory,
            short_term_memory=short_term_memory,
            long_term_memory=long_term_memory,
            **kwargs
        )

    @classmethod
    def from_vstore(
        cls,
        vector_store_connector: "VectorStoreConnector",
        embeddings: Optional[Embeddings] = None,
        executor: Optional[Executor] = None,
        now: Optional[datetime] = None,
        sensory_memory: Optional[SensoryMemory[T]] = None,
        short_term_memory: Optional[ShortTermMemory[T]] = None,
        long_term_memory: Optional[LongTermMemory[T]] = None,
        **kwargs
    ):
        """Create a hybrid memory from vector store."""
        if not embeddings:
            embeddings = vector_store_connector.current_embeddings
        if not executor:
            executor = ThreadPoolExecutor()
        if not now:
            now = datetime.now()

        if not sensory_memory:
            sensory_memory = SensoryMemory()
        if not short_term_memory:
            if not embeddings:
                raise ValueError("embeddings is required.")
            short_term_memory = EnhancedShortTermMemory(embeddings, executor)
        if not long_term_memory:
            long_term_memory = LongTermMemory(
                executor,
                vector_store_connector,
                now=now,
            )
        return cls(now, sensory_memory, short_term_memory, long_term_memory, **kwargs)

    def initialize(
        self,
        name: Optional[str] = None,
        llm_client: Optional[LLMClient] = None,
        importance_scorer: Optional[ImportanceScorer[T]] = None,
        insight_extractor: Optional[InsightExtractor[T]] = None,
        real_memory_fragment_class: Optional[Type[T]] = None,
    ) -> None:
        """Initialize the memory.

        It will initialize all the memories.
        """
        memories = [
            self._sensory_memory,
            self._short_term_memory,
            self._long_term_memory,
        ]
        kwargs = {
            "name": name,
            "llm_client": llm_client,
            "importance_scorer": importance_scorer,
            "insight_extractor": insight_extractor,
            "real_memory_fragment_class": real_memory_fragment_class,
        }
        for memory in memories:
            memory.initialize(**kwargs)
        super().initialize(**kwargs)

    @mutable
    async def write(
        self,
        memory_fragment: T,
        now: Optional[datetime] = None,
        op: WriteOperation = WriteOperation.ADD,
    ) -> Optional[DiscardedMemoryFragments[T]]:
        """Write a memory fragment to the memory."""
        # First write to sensory memory
        sen_discarded_memories = await self._sensory_memory.write(memory_fragment)
        if not sen_discarded_memories:
            return None
        short_term_discarded_memories = []
        discarded_memory_fragments = []
        discarded_insights = []
        for sen_memory in sen_discarded_memories.discarded_memory_fragments:
            # Write to short term memory
            short_discarded_memory = await self._short_term_memory.write(sen_memory)
            if short_discarded_memory:
                short_term_discarded_memories.append(short_discarded_memory)
                discarded_memory_fragments.extend(
                    short_discarded_memory.discarded_memory_fragments
                )
                for insight in short_discarded_memory.discarded_insights:
                    # Just keep the first insight
                    discarded_insights.append(insight.insights[0])
        # Obtain the importance of insights
        insight_scores = await self.score_memory_importance(discarded_insights)
        # Get the importance of insights
        for i, ins in enumerate(discarded_insights):
            ins.update_importance(insight_scores[i])
        all_memories = discarded_memory_fragments + discarded_insights
        if self._long_term_memory:
            # Write to long term memory
            await self._long_term_memory.write_batch(all_memories, self.now)
        return None

    @immutable
    async def read(
        self,
        observation: str,
        alpha: Optional[float] = None,
        beta: Optional[float] = None,
        gamma: Optional[float] = None,
    ) -> List[T]:
        """Read memories from the memory."""
        (
            retrieved_long_term_memories,
            short_term_discarded_memories,
        ) = await self.fetch_memories(observation, self._short_term_memory)

        await self.save_memories_after_retrieval(short_term_discarded_memories)
        return retrieved_long_term_memories

    @immutable
    async def fetch_memories(
        self,
        observation: str,
        short_term_memory: Optional[ShortTermMemory[T]] = None,
    ) -> Tuple[List[T], List[DiscardedMemoryFragments[T]]]:
        """Fetch memories from long term memory.

        If short_term_memory is provided, write the fetched memories to the short term
        memory.
        """
        retrieved_long_term_memories = await self._long_term_memory.fetch_memories(
            observation
        )
        if not short_term_memory:
            return retrieved_long_term_memories, []
        short_term_discarded_memories: List[DiscardedMemoryFragments[T]] = []
        discarded_memory_fragments: List[T] = []
        for ltm in retrieved_long_term_memories:
            short_discarded_memory = await short_term_memory.write(
                ltm, op=WriteOperation.RETRIEVAL
            )
            if short_discarded_memory:
                short_term_discarded_memories.append(short_discarded_memory)
                discarded_memory_fragments.extend(
                    short_discarded_memory.discarded_memory_fragments
                )
        for stm in short_term_memory.short_term_memories:
            retrieved_long_term_memories.append(
                stm.current_class.build_from(
                    observation=stm.raw_observation,
                    importance=stm.importance,
                )
            )
        return retrieved_long_term_memories, short_term_discarded_memories

    async def save_memories_after_retrieval(
        self, fragments: List[DiscardedMemoryFragments[T]]
    ):
        """Save memories after retrieval."""
        discarded_memory_fragments = []
        discarded_memory_insights: List[T] = []
        for f in fragments:
            discarded_memory_fragments.extend(f.discarded_memory_fragments)
            for fi in f.discarded_insights:
                discarded_memory_insights.append(fi.insights[0])
        insights_importance = await self.score_memory_importance(
            discarded_memory_insights
        )
        for i, ins in enumerate(discarded_memory_insights):
            ins.update_importance(insights_importance[i])
        all_memories = discarded_memory_fragments + discarded_memory_insights
        await self._long_term_memory.write_batch(all_memories, self.now)

    async def clear(self) -> List[T]:
        """Clear the memory.

        # TODO
        """
        return []



"""Code visualization protocol."""
from ..base import Vis


class VisCode(Vis):
    """Protocol for visualizing code."""

    @classmethod
    def vis_tag(cls) -> str:
        """Return the tag name of the vis protocol module."""
        return "vis-code"



"""The AWEL Agent Operator Resource."""

from typing import Any, Dict, List, Optional

from dbgpt._private.pydantic import BaseModel, ConfigDict, Field, model_validator
from dbgpt.core import LLMClient
from dbgpt.core.awel.flow import (
    FunctionDynamicOptions,
    OptionValue,
    Parameter,
    ResourceCategory,
    register_resource,
)

from ....resource.base import AgentResource
from ....resource.manage import get_resource_manager
from ....util.llm.llm import LLMConfig, LLMStrategyType
from ...agent_manage import get_agent_manager


def _load_resource_types():
    resources = get_resource_manager().get_supported_resources()
    return [OptionValue(label=item, name=item, value=item) for item in resources.keys()]


@register_resource(
    label="AWEL Agent Resource",
    name="agent_operator_resource",
    description="The Agent Resource.",
    category=ResourceCategory.AGENT,
    parameters=[
        Parameter.build_from(
            label="Agent Resource Type",
            name="agent_resource_type",
            type=str,
            optional=True,
            default=None,
            options=FunctionDynamicOptions(func=_load_resource_types),
        ),
        Parameter.build_from(
            label="Agent Resource Name",
            name="agent_resource_name",
            type=str,
            optional=True,
            default=None,
            description="The agent resource name.",
        ),
        Parameter.build_from(
            label="Agent Resource Value",
            name="agent_resource_value",
            type=str,
            optional=True,
            default=None,
            description="The agent resource value.",
        ),
    ],
    alias=[
        "dbgpt.serve.agent.team.layout.agent_operator_resource.AwelAgentResource",
        "dbgpt.agent.plan.awel.agent_operator_resource.AWELAgentResource",
    ],
)
class AWELAgentResource(AgentResource):
    """AWEL Agent Resource."""

    @model_validator(mode="before")
    @classmethod
    def pre_fill(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Pre fill the agent ResourceType."""
        if not isinstance(values, dict):
            return values
        name = values.pop("agent_resource_name")
        type = values.pop("agent_resource_type")
        value = values.pop("agent_resource_value")

        values["name"] = name
        values["type"] = type
        values["value"] = value

        return values


@register_resource(
    label="AWEL Agent LLM Config",
    name="agent_operator_llm_config",
    description="The Agent LLM Config.",
    category=ResourceCategory.AGENT,
    parameters=[
        Parameter.build_from(
            "LLM Client",
            "llm_client",
            LLMClient,
            optional=True,
            default=None,
            description="The LLM Client.",
        ),
        Parameter.build_from(
            label="Agent LLM Strategy",
            name="llm_strategy",
            type=str,
            optional=True,
            default=None,
            options=[
                OptionValue(label=item.name, name=item.value, value=item.value)
                for item in LLMStrategyType
            ],
            description="The Agent LLM Strategy.",
        ),
        Parameter.build_from(
            label="Agent LLM Strategy Value",
            name="strategy_context",
            type=str,
            optional=True,
            default=None,
            description="The agent LLM Strategy Value.",
        ),
    ],
    alias=[
        "dbgpt.serve.agent.team.layout.agent_operator_resource.AwelAgentConfig",
        "dbgpt.agent.plan.awel.agent_operator_resource.AWELAgentConfig",
    ],
)
class AWELAgentConfig(LLMConfig):
    """AWEL Agent Config."""

    pass


def _agent_resource_option_values() -> List[OptionValue]:
    return [
        OptionValue(label=item["name"], name=item["name"], value=item["name"])
        for item in get_agent_manager().list_agents()
    ]


@register_resource(
    label="AWEL Layout Agent",
    name="agent_operator_agent",
    description="The Agent to build the Agent Operator.",
    category=ResourceCategory.AGENT,
    parameters=[
        Parameter.build_from(
            label="Agent Profile",
            name="agent_profile",
            type=str,
            description="Which agent want use.",
            options=FunctionDynamicOptions(func=_agent_resource_option_values),
        ),
        Parameter.build_from(
            label="Role Name",
            name="role_name",
            type=str,
            optional=True,
            default=None,
            description="The agent role name.",
        ),
        Parameter.build_from(
            label="Fixed Gogal",
            name="fixed_subgoal",
            type=str,
            optional=True,
            default=None,
            description="The agent fixed gogal.",
        ),
        Parameter.build_from(
            label="Agent Resource",
            name="agent_resource",
            type=AWELAgentResource,
            optional=True,
            default=None,
            description="The agent resource.",
        ),
        Parameter.build_from(
            label="Agent LLM  Config",
            name="agent_llm_Config",
            type=AWELAgentConfig,
            optional=True,
            default=None,
            description="The agent llm config.",
        ),
    ],
    alias=[
        "dbgpt.serve.agent.team.layout.agent_operator_resource.AwelAgent",
        "dbgpt.agent.plan.awel.agent_operator_resource.AWELAgent",
    ],
)
class AWELAgent(BaseModel):
    """AWEL Agent."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    agent_profile: str
    role_name: Optional[str] = None
    llm_config: Optional[LLMConfig] = None
    resources: List[AgentResource] = Field(default_factory=list)
    fixed_subgoal: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def pre_fill(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Pre fill the agent ResourceType."""
        if not isinstance(values, dict):
            return values
        resource = values.pop("agent_resource")
        llm_config = values.pop("agent_llm_Config")

        if resource is not None:
            values["resources"] = [resource]

        if llm_config is not None:
            values["llm_config"] = llm_config

        return values



"""Base class for all trigger classes."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic

from ..operators.common_operator import TriggerOperator
from ..task.base import OUT


class Trigger(TriggerOperator[OUT], ABC, Generic[OUT]):
    """Base class for all trigger classes.

    Now only support http trigger.
    """

    @abstractmethod
    async def trigger(self, **kwargs) -> Any:
        """Trigger the workflow or a specific operation in the workflow."""



"""Transformer base class."""
import logging
from abc import ABC, abstractmethod
from typing import List, Optional

logger = logging.getLogger(__name__)


class TransformerBase:
    """Transformer base class."""


class EmbedderBase(TransformerBase, ABC):
    """Embedder base class."""


class ExtractorBase(TransformerBase, ABC):
    """Extractor base class."""

    @abstractmethod
    async def extract(self, text: str, limit: Optional[int] = None) -> List:
        """Extract results from text."""


class TranslatorBase(TransformerBase, ABC):
    """Translator base class."""



"""GPTs memory."""

import json
from collections import OrderedDict, defaultdict
from typing import Dict, List, Optional

from dbgpt.vis.client import VisAgentMessages, VisAgentPlans, vis_client

from ...action.base import ActionOutput
from .base import GptsMessage, GptsMessageMemory, GptsPlansMemory
from .default_gpts_memory import DefaultGptsMessageMemory, DefaultGptsPlansMemory

NONE_GOAL_PREFIX: str = "none_goal_count_"


class GptsMemory:
    """GPTs memory."""

    def __init__(
        self,
        plans_memory: Optional[GptsPlansMemory] = None,
        message_memory: Optional[GptsMessageMemory] = None,
    ):
        """Create a memory to store plans and messages."""
        self._plans_memory: GptsPlansMemory = (
            plans_memory if plans_memory is not None else DefaultGptsPlansMemory()
        )
        self._message_memory: GptsMessageMemory = (
            message_memory if message_memory is not None else DefaultGptsMessageMemory()
        )

    @property
    def plans_memory(self) -> GptsPlansMemory:
        """Return the plans memory."""
        return self._plans_memory

    @property
    def message_memory(self) -> GptsMessageMemory:
        """Return the message memory."""
        return self._message_memory

    async def _message_group_vis_build(self, message_group):
        if not message_group:
            return ""
        num: int = 0
        last_goal = next(reversed(message_group))
        last_goal_messages = message_group[last_goal]

        last_goal_message = last_goal_messages[-1]
        vis_items = []

        plan_temps = []
        for key, value in message_group.items():
            num = num + 1
            if key.startswith(NONE_GOAL_PREFIX):
                vis_items.append(await self._messages_to_plan_vis(plan_temps))
                plan_temps = []
                num = 0
                vis_items.append(await self._messages_to_agents_vis(value))
            else:
                num += 1
                plan_temps.append(
                    {
                        "name": key,
                        "num": num,
                        "status": "complete",
                        "agent": value[0].receiver if value else "",
                        "markdown": await self._messages_to_agents_vis(value),
                    }
                )

        if len(plan_temps) > 0:
            vis_items.append(await self._messages_to_plan_vis(plan_temps))
        vis_items.append(await self._messages_to_agents_vis([last_goal_message]))
        return "\n".join(vis_items)

    async def _plan_vis_build(self, plan_group: dict[str, list]):
        num: int = 0
        plan_items = []
        for key, value in plan_group.items():
            num = num + 1
            plan_items.append(
                {
                    "name": key,
                    "num": num,
                    "status": "complete",
                    "agent": value[0].receiver if value else "",
                    "markdown": await self._messages_to_agents_vis(value),
                }
            )
        return await self._messages_to_plan_vis(plan_items)

    async def one_chat_completions_v2(self, conv_id: str):
        """Generate a visualization of the conversation."""
        messages = self.message_memory.get_by_conv_id(conv_id=conv_id)
        temp_group: Dict[str, List[GptsMessage]] = OrderedDict()
        none_goal_count = 1
        count: int = 0
        for message in messages:
            count = count + 1
            if count == 1:
                continue
            current_goal = message.current_goal

            last_goal = next(reversed(temp_group)) if temp_group else None
            if last_goal:
                last_goal_messages = temp_group[last_goal]
                if current_goal:
                    if current_goal == last_goal:
                        last_goal_messages.append(message)
                    else:
                        temp_group[current_goal] = [message]
                else:
                    temp_group[f"{NONE_GOAL_PREFIX}{none_goal_count}"] = [message]
                    none_goal_count += 1
            else:
                if current_goal:
                    temp_group[current_goal] = [message]
                else:
                    temp_group[f"{NONE_GOAL_PREFIX}{none_goal_count}"] = [message]
                    none_goal_count += 1

        return await self._message_group_vis_build(temp_group)

    async def one_chat_completions(self, conv_id: str):
        """Generate a visualization of the conversation."""
        messages = self.message_memory.get_by_conv_id(conv_id=conv_id)
        temp_group: Dict[str, List[GptsMessage]] = defaultdict(list)
        temp_messages = []
        vis_items = []
        count: int = 0
        for message in messages:
            count = count + 1
            if count == 1:
                continue
            if not message.current_goal or len(message.current_goal) <= 0:
                if len(temp_group) > 0:
                    vis_items.append(await self._plan_vis_build(temp_group))
                    temp_group.clear()

                temp_messages.append(message)
            else:
                if len(temp_messages) > 0:
                    vis_items.append(await self._messages_to_agents_vis(temp_messages))
                    temp_messages.clear()

                last_goal = message.current_goal
                temp_group[last_goal].append(message)

        if len(temp_group) > 0:
            vis_items.append(await self._plan_vis_build(temp_group))
            temp_group.clear()
        if len(temp_messages) > 0:
            vis_items.append(await self._messages_to_agents_vis(temp_messages, True))
            temp_messages.clear()

        return "\n".join(vis_items)

    async def _messages_to_agents_vis(
        self, messages: List[GptsMessage], is_last_message: bool = False
    ):
        if messages is None or len(messages) <= 0:
            return ""
        messages_view = []
        for message in messages:
            action_report_str = message.action_report
            view_info = message.content
            if action_report_str and len(action_report_str) > 0:
                action_out = ActionOutput.from_dict(json.loads(action_report_str))
                if action_out is not None and (
                    action_out.is_exe_success or is_last_message
                ):
                    view = action_out.view
                    view_info = view if view else action_out.content

            messages_view.append(
                {
                    "sender": message.sender,
                    "receiver": message.receiver,
                    "model": message.model_name,
                    "markdown": view_info,
                }
            )
        vis_compent = vis_client.get(VisAgentMessages.vis_tag())
        return await vis_compent.display(content=messages_view)

    async def _messages_to_plan_vis(self, messages: List[Dict]):
        if messages is None or len(messages) <= 0:
            return ""
        return await vis_client.get(VisAgentPlans.vis_tag()).display(content=messages)



"""Memory module for GPTS messages and plans.

It stores the messages and plans generated of multiple agents in the conversation.

It is different from the agent memory as it is a formatted structure to store the
messages and plans, and it can be stored in a database or a file.
"""

from .base import (  # noqa: F401
    GptsMessage,
    GptsMessageMemory,
    GptsPlan,
    GptsPlansMemory,
)
from .default_gpts_memory import (  # noqa: F401
    DefaultGptsMessageMemory,
    DefaultGptsPlansMemory,
)
from .gpts_memory import GptsMemory  # noqa: F401
